from __future__ import unicode_literals

import datetime
import itertools

from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.validators import MinValueValidator, MaxValueValidator
from django.db import models
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.lru_cache import lru_cache
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker

from nodeconductor.core.fields import JSONField
from nodeconductor.core import models as core_models, utils as core_utils
from nodeconductor.core.exceptions import IncorrectStateException
from nodeconductor.structure import models as structure_models

from nodeconductor_assembly_waldur.common import mixins as common_mixins
from nodeconductor_assembly_waldur.packages import models as package_models
from nodeconductor_assembly_waldur.support import models as support_models

from . import managers, utils, registrators


@python_2_unicode_compatible
class Invoice(core_models.UuidMixin, models.Model):
    """ Invoice describes billing information about purchased packages for customers on a monthly basis """

    class Permissions(object):
        customer_path = 'customer'

    class Meta(object):
        unique_together = ('customer', 'month', 'year')

    class States(object):
        PENDING = 'pending'
        CREATED = 'created'
        PAID = 'paid'
        CANCELED = 'canceled'

        CHOICES = ((PENDING, _('Pending')), (CREATED, _('Created')), (PAID, _('Paid')), (CANCELED, _('Canceled')))

    month = models.PositiveSmallIntegerField(default=utils.get_current_month,
                                             validators=[MinValueValidator(1), MaxValueValidator(12)])
    year = models.PositiveSmallIntegerField(default=utils.get_current_year)
    state = models.CharField(max_length=30, choices=States.CHOICES, default=States.PENDING)
    customer = models.ForeignKey(structure_models.Customer, verbose_name=_('organization'), related_name='+')
    tax_percent = models.DecimalField(default=0, max_digits=4, decimal_places=2,
                                      validators=[MinValueValidator(0), MaxValueValidator(100)])
    invoice_date = models.DateField(null=True, blank=True,
                                    help_text=_('Date then invoice moved from state pending to created.'))

    tracker = FieldTracker()

    @property
    def tax(self):
        return self.price * self.tax_percent / 100

    @property
    def total(self):
        return self.price + self.tax

    @property
    def price(self):
        return sum((item.price for item in self.items))

    @property
    def items(self):
        return itertools.chain(self.openstack_items.all(),
                               self.offering_items.all(),
                               self.generic_items.all())

    @property
    def due_date(self):
        if self.invoice_date:
            return self.invoice_date + datetime.timedelta(days=settings.INVOICES['PAYMENT_INTERVAL'])

    @property
    def number(self):
        return 100000 + self.id

    def set_created(self):
        """
        Performs following actions:
            - Freeze all invoice items
            - Change state from pending to billed
        """
        if self.state != self.States.PENDING:
            raise IncorrectStateException(_('Invoice must be in pending state.'))

        self.state = self.States.CREATED
        self.invoice_date = timezone.now().date()
        self.save(update_fields=['state', 'invoice_date'])

    def freeze(self):
        for item in self.items:
            item.freeze()

    def register_offering(self, offering, start=None):
        if start is None:
            start = timezone.now()

        end = core_utils.month_end(start)
        OfferingItem.objects.create(
            offering=offering,
            unit_price=offering.unit_price,
            unit=offering.unit,
            invoice=self,
            start=start,
            end=end,
        )

    def __str__(self):
        return '%s | %s-%s' % (self.customer, self.year, self.month)


@python_2_unicode_compatible
class InvoiceItem(common_mixins.ProductCodeMixin, common_mixins.UnitPriceMixin):
    """
    Mixin which identifies invoice item to be used for price calculation.
    """

    class Meta(object):
        abstract = True

    start = models.DateTimeField(default=utils.get_current_month_start,
                                 help_text=_('Date and time when item usage has started.'))
    end = models.DateTimeField(default=utils.get_current_month_end,
                               help_text=_('Date and time when item usage has ended.'))

    # Project name and UUID should be stored separately because project is not available after removal
    project = models.ForeignKey(structure_models.Project, on_delete=models.SET_NULL, null=True)
    project_name = models.CharField(max_length=150, blank=True)
    project_uuid = models.CharField(max_length=32, blank=True)

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]

    @property
    def tax(self):
        return self.price * self.invoice.tax_percent / 100

    @property
    def total(self):
        return self.price + self.tax

    @property
    def price(self):
        if self.unit == self.Units.QUANTITY:
            return self.unit_price * self.quantity
        elif self.unit == self.Units.PER_DAY:
            return self.unit_price * self.usage_days
        elif self.unit == self.Units.PER_HALF_MONTH:
            if self.start.day >= 15 or self.end.day <= 15:
                return self.unit_price
            else:
                return self.unit_price * 2

        return self.unit_price

    @property
    def usage_days(self):
        """
        Returns the number of days package was used from the time
        it was purchased or from the start of current month
        """
        full_days = utils.get_full_days(self.start, self.end)
        return full_days

    def terminate(self, end=None):
        self.freeze()
        self.end = end or timezone.now()
        self.save(update_fields=['end'])

    def name(self):
        raise NotImplementedError()

    def freeze(self):
        raise NotImplementedError()

    def __str__(self):
        return self.name


class GenericInvoiceItem(InvoiceItem):
    invoice = models.ForeignKey(Invoice, related_name='generic_items')
    content_type = models.ForeignKey(ContentType, null=True, related_name='+')
    object_id = models.PositiveIntegerField(null=True)
    quantity = models.PositiveIntegerField(default=0)

    scope = GenericForeignKey('content_type', 'object_id')
    details = JSONField(default={}, blank=True, help_text=_('Stores data about scope'))

    objects = managers.GenericInvoiceItemManager()
    tracker = FieldTracker()

    @property
    def name(self):
        if self.details:
            return self.details['name']
        return registrators.RegistrationManager.get_name(self.scope)

    def freeze(self):
        if self.scope:
            self.details = registrators.RegistrationManager.get_details(self.scope)
            self.details['name'] = registrators.RegistrationManager.get_name(self.scope)
            self.details['scope_uuid'] = self.scope.uuid.hex
            self.save(update_fields=['details'])


class OfferingItem(InvoiceItem):
    """ OfferingItem stores details for invoices about purchased custom offering item. """
    invoice = models.ForeignKey(Invoice, related_name='offering_items')
    offering = models.ForeignKey(support_models.Offering, on_delete=models.SET_NULL, null=True, related_name='+')
    offering_details = JSONField(default={}, blank=True, help_text=_('Stores data about offering'))
    tracker = FieldTracker()

    @property
    def name(self):
        if self.offering_details:
            return '%s (%s)' % (self.project_name, self.offering_details['offering_type'])

        return '%s (%s)' % (self.offering.project.name, self.offering.type)

    def freeze(self):
        """
        Saves offering type and project name in "offering_details" if offering exists
        """
        if self.offering:
            self.offering_details['offering_type'] = self.offering.type
            self.offering_details['offering_uuid'] = self.offering.uuid.hex
            self.save(update_fields=['offering_details'])

    def get_offering_uuid(self):
        if self.offering:
            return self.offering.uuid.hex
        else:
            return self.offering_details.get('offering_uuid')

    def get_offering_type(self):
        if self.offering:
            return self.offering.type
        else:
            return self.offering_details.get('offering_type')


class OpenStackItem(InvoiceItem):
    """ OpenStackItem stores details for invoices about purchased OpenStack packages """

    invoice = models.ForeignKey(Invoice, related_name='openstack_items')

    package = models.ForeignKey(package_models.OpenStackPackage, on_delete=models.SET_NULL, null=True, related_name='+')
    package_details = JSONField(default={}, blank=True, help_text=_('Stores data about package'))
    tracker = FieldTracker()

    @property
    def name(self):
        template_category = self.get_template_category()
        if template_category:
            return '%s (%s / %s)' % (self.get_tenant_name(), template_category, self.get_template_name())
        else:
            return '%s (%s)' % (self.get_tenant_name(), self.get_template_name())

    def freeze(self):
        """
        Saves tenant and package template names and uuids in "package_details" if package exists
        """
        if self.package:
            self.package_details['tenant_name'] = self.package.tenant.name
            self.package_details['tenant_uuid'] = self.package.tenant.uuid.hex
            self.package_details['template_name'] = self.package.template.name
            self.package_details['template_uuid'] = self.package.template.uuid.hex
            self.package_details['template_category'] = self.package.template.category
            self.save(update_fields=['package_details'])

    def shift_backward(self, days=1):
        """
        Shifts end date to N 'days' ago.
        If N is larger than it lasts - zero length will be set.
        :param days: number of days to shift end date
        """
        if (self.end - self.start).days > days:
            end = self.end - timezone.timedelta(days=1)
        else:
            end = self.start

        self.end = end
        self.save()

    def extend_to_the_end_of_the_day(self):
        self.end = self.end.replace(hour=23, minute=59, second=59)
        self.save()

    def get_tenant_name(self):
        if self.package:
            return self.package.tenant.name
        else:
            return self.package_details.get('tenant_name')

    def get_tenant_uuid(self):
        if self.package:
            return self.package.tenant.uuid.hex
        else:
            return self.package_details.get('tenant_uuid')

    def get_template_name(self):
        if self.package:
            return self.package.template.name
        else:
            return self.package_details.get('template_name')

    def get_template_uuid(self):
        if self.package:
            return self.package.template.uuid.hex
        else:
            return self.package_details.get('template_uuid')

    def get_template_category(self):
        if self.package:
            return self.package.template.get_category_display()
        else:
            template_category = self.package_details.get('template_category')
            if template_category:
                categories = dict(package_models.PackageTemplate.Categories.CHOICES)
                template_category = categories[template_category]

            return template_category


@python_2_unicode_compatible
class PaymentDetails(core_models.UuidMixin, models.Model):
    """ Customer payment details """

    class Permissions(object):
        customer_path = 'customer'

    class Meta(object):
        verbose_name = _('Payment details')
        verbose_name_plural = _('Payment details')

    customer = models.OneToOneField(structure_models.Customer, related_name='payment_details')
    accounting_start_date = models.DateTimeField(_('Start date of accounting'), default=timezone.now)
    company = models.CharField(blank=True, max_length=150)
    type = models.CharField(blank=True, max_length=150)
    address = models.CharField(blank=True, max_length=300)
    country = models.CharField(blank=True, max_length=50)
    email = models.EmailField(blank=True, max_length=75)
    postal = models.CharField(blank=True, max_length=20)
    phone = models.CharField(blank=True, max_length=20)
    bank = models.CharField(blank=True, max_length=150)
    account = models.CharField(blank=True, max_length=50)
    default_tax_percent = models.DecimalField(default=0, max_digits=4, decimal_places=2,
                                              validators=[MinValueValidator(0), MaxValueValidator(100)])

    @classmethod
    def get_url_name(cls):
        return 'payment-details'

    def is_billable(self):
        return timezone.now() >= self.accounting_start_date

    def __str__(self):
        return 'PaymentDetails for %s' % self.customer



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


def set_tenant_extra_configuration(apps, schema_editor):
    OpenStackPackage = apps.get_model('packages', 'OpenStackPackage')

    for package in OpenStackPackage.objects.all():
        package.tenant.extra_configuration = {
            'package_name': package.template.name,
            'package_uuid': package.template.uuid.hex,
            'package_category': package.template.get_category_display(),
        }
        package.tenant.save()


class Migration(migrations.Migration):

    dependencies = [
        ('packages', '0008_package_component_type'),
    ]

    operations = [
        migrations.RunPython(set_tenant_extra_configuration),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import nodeconductor.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0022_volume_device'),
        ('structure', '0037_remove_customer_billing_backend_id'),
        ('packages', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='OpenStackPackage',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', nodeconductor.core.fields.UUIDField()),
                ('service_settings', models.ForeignKey(related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.RemoveField(
            model_name='packagetemplate',
            name='type',
        ),
        migrations.AddField(
            model_name='packagetemplate',
            name='service_settings',
            field=models.ForeignKey(related_name='+', to='structure.ServiceSettings'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='openstackpackage',
            name='template',
            field=models.ForeignKey(related_name='openstack_packages', to='packages.PackageTemplate', help_text='Tenant will be created based on this template.'),
        ),
        migrations.AddField(
            model_name='openstackpackage',
            name='tenant',
            field=models.ForeignKey(related_name='+', to='openstack.Tenant'),
        ),
    ]



from __future__ import unicode_literals

from django import forms
from django.contrib import admin
from django.contrib.auth import get_user_model

from nodeconductor.structure import admin as structure_admin

from . import models


User = get_user_model()


class UserChoiceField(forms.ModelChoiceField):
    def label_from_instance(self, user):
        return '{} - {}'.format(user.full_name, user.username)


class SupportUserAdminForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super(SupportUserAdminForm, self).__init__(*args, **kwargs)
        self.fields['user'] = UserChoiceField(queryset=User.objects.all().order_by('full_name'))


class SupportUserAdmin(admin.ModelAdmin):
    form = SupportUserAdminForm


class OfferingAdmin(admin.ModelAdmin):
    list_display = ('type', 'name', 'unit_price', 'unit', 'state')
    fields = ('name', 'unit_price', 'unit', 'type', 'issue',
              'project', 'state', 'product_code', 'article_code')


class IssueAdmin(structure_admin.BackendModelAdmin):
    exclude = ('resource_content_type', 'resource_object_id')


admin.site.register(models.Offering, OfferingAdmin)
admin.site.register(models.Issue, IssueAdmin)
admin.site.register(models.Comment, structure_admin.BackendModelAdmin)
admin.site.register(models.SupportUser, SupportUserAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


def add_openstack_component_details_to_tenant(apps, schema_editor):
    OpenStackPackage = apps.get_model('packages', 'OpenStackPackage')

    for package in OpenStackPackage.objects.all():
        for component in package.template.components.all():
            package.tenant.extra_configuration[component.type] = component.amount

        package.tenant.save()


class Migration(migrations.Migration):

    dependencies = [
        ('packages', '0010_packagetemplate_protect'),
    ]

    operations = [
        migrations.RunPython(add_openstack_component_details_to_tenant),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import model_utils.fields
import nodeconductor.core.fields
import django.utils.timezone
from decimal import Decimal
import django.db.models.deletion
import django.core.validators
import nodeconductor.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0039_remove_permission_groups'),
        ('support', '0005_issue_first_response_sla'),
    ]

    operations = [
        migrations.CreateModel(
            name='Offering',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[nodeconductor.core.validators.validate_name])),
                ('uuid', nodeconductor.core.fields.UUIDField()),
                ('type', models.CharField(max_length=255, blank=True)),
                ('type_label', models.CharField(max_length=255, blank=True)),
                ('price', models.DecimalField(decimal_places=7, default=0, max_digits=13, validators=[django.core.validators.MinValueValidator(Decimal('0'))], help_text='The price per unit of offering', verbose_name='Price per day')),
                ('state', models.CharField(default='requested', max_length=30, choices=[('requested', 'Requested'), ('ok', 'OK'), ('terminated', 'Terminated')])),
                ('issue', models.ForeignKey(on_delete=django.db.models.deletion.SET_NULL, to='support.Issue', null=True)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='structure.Project', null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.3 on 2017-09-20 06:45
from __future__ import unicode_literals

from decimal import Decimal
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('experts', '0006_expertrequest_issue'),
    ]

    operations = [
        migrations.AddField(
            model_name='expertrequest',
            name='article_code',
            field=models.CharField(blank=True, max_length=30),
        ),
        migrations.AddField(
            model_name='expertrequest',
            name='common_tos',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='expertrequest',
            name='contract_methodology',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='expertrequest',
            name='milestones',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='expertrequest',
            name='objectives',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='expertrequest',
            name='out_of_scope',
            field=models.TextField(blank=True, help_text='Elements that are explicitly excluded from the contract'),
        ),
        migrations.AddField(
            model_name='expertrequest',
            name='price',
            field=models.DecimalField(decimal_places=7, default=0, max_digits=22, validators=[django.core.validators.MinValueValidator(Decimal('0'))]),
        ),
        migrations.AddField(
            model_name='expertrequest',
            name='product_code',
            field=models.CharField(blank=True, max_length=30),
        ),
        migrations.AddField(
            model_name='expertrequest',
            name='recurring_billing',
            field=models.BooleanField(default=False, help_text='Defines whether expert request has to be billed every month or only once'),
        ),
        migrations.AlterField(
            model_name='expertcontract',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='+', to='structure.Project'),
        ),
    ]



from nodeconductor_assembly_waldur.invoices import models
from nodeconductor_assembly_waldur.invoices.registrators import BaseRegistrator
from nodeconductor_assembly_waldur.support import models as support_models


class OfferingItemRegistrator(BaseRegistrator):

    def get_sources(self, customer):
        return support_models.Offering.objects.filter(project__customer=customer).distinct()

    def has_sources(self, customer):
        return self.get_sources(customer).exists()

    def get_customer(self, source):
        return source.project.customer

    def _find_item(self, source, now):
        offering = source
        result = models.OfferingItem.objects.filter(
            offering=offering,
            invoice__customer=offering.project.customer,
            invoice__state=models.Invoice.States.PENDING,
            invoice__year=now.year,
            invoice__month=now.month,
        ).first()
        return result

    def _create_item(self, source, invoice, start, end):
        offering = source
        result = models.OfferingItem.objects.create(
            offering=offering,
            project=offering.project,
            unit_price=offering.unit_price,
            unit=offering.unit,
            product_code=offering.product_code,
            article_code=offering.article_code,
            invoice=invoice,
            start=start,
            end=end,
        )
        return result



from __future__ import unicode_literals

import logging

from celery import shared_task
from django.conf import settings
from django.core.mail import send_mail
from django.template.loader import render_to_string

from . import models

logger = logging.getLogger(__name__)


@shared_task(name='nodeconductor_assembly_waldur.experts.send_contract')
def send_contract(request_uuid, email):
    request = models.ExpertRequest.objects.get(uuid=request_uuid)

    context = dict(
        request=request,
        customer_name=request.project.customer.name,
        project_name=request.project.name,
    )

    subject = render_to_string('experts/contract_subject.txt', context).strip()
    text_message = render_to_string('experts/contract_message.txt', context)
    html_message = render_to_string('experts/contract_message.html', context)

    logger.debug('About to send expert contract {request_name} to {email}.'.format(
        request_name=request.name, email=email, **context))
    send_mail(subject, text_message, settings.DEFAULT_FROM_EMAIL, [email], html_message=html_message)



import logging

from smtplib import SMTPException

from celery import Task, shared_task
from django.conf import settings
from django.core.mail import send_mail
from django.template.loader import render_to_string

from nodeconductor.core import utils as core_utils


from . import backend, models


logger = logging.getLogger(__name__)


class SupportUserPullTask(Task):
    """ Pull support users from backend """
    name = 'support.SupportUserPullTask'

    def run(self):
        if not settings.WALDUR_SUPPORT['ENABLED']:
            return

        backend_users = backend.get_active_backend().get_users()
        for backend_user in backend_users:
            user, created = models.SupportUser.objects.get_or_create(
                backend_id=backend_user.backend_id, defaults={'name': backend_user.name})
            if not created and user.name != backend_user.name:
                user.name = backend_user.name
                user.save()
        models.SupportUser.objects.exclude(backend_id__in=[u.backend_id for u in backend_users]).delete()


@shared_task(name='nodeconductor_assembly_waldur.support.send_issue_updated_notification')
def send_issue_updated_notification(serialized_issue):
    issue = core_utils.deserialize_instance(serialized_issue)
    _send_issue_notification(issue, 'issue_updated')


@shared_task(name='nodeconductor_assembly_waldur.support.send_comment_added_notification')
def send_comment_added_notification(serialized_issue):
    issue = core_utils.deserialize_instance(serialized_issue)
    _send_issue_notification(issue, 'comment_added')


def _send_issue_notification(issue, template):
    if not settings.WALDUR_SUPPORT['ENABLED']:
        return

    if settings.SUPPRESS_NOTIFICATION_EMAILS:
        message = ('Issue notifications are suppressed. '
                   'Please set SUPPRESS_NOTIFICATION_EMAILS to False to send notifications.')
        logger.info(message)
        return

    receiver = issue.caller

    context = {
        'issue_url': settings.ISSUE_LINK_TEMPLATE.format(uuid=issue.uuid)
    }

    subject = render_to_string('support/notification_%s_subject.txt' % template).strip()
    text_message = render_to_string('support/notification_%s.txt' % template, context)
    html_message = render_to_string('support/notification_%s.html' % template, context)

    logger.debug('About to send an issue update notification to %s' % receiver.email)

    try:
        send_mail(subject, text_message, settings.DEFAULT_FROM_EMAIL, [receiver.email], html_message=html_message)
    except SMTPException as e:
        message = 'Failed to notify a user about an issue update. Issue uuid: %s. Error: %s' % (issue.uuid, e.message)
        logger.warning(message)



import importlib

from django.conf import settings


def get_active_backend():
    path = settings.WALDUR_SUPPORT['ACTIVE_BACKEND']
    module_path, class_name = path.split(':')
    module = importlib.import_module(module_path)
    klass = getattr(module, class_name)
    return klass()


class SupportBackendError(Exception):
    pass


class SupportBackend(object):
    """ Interface for support backend """
    def create_issue(self, issue):
        pass

    def update_issue(self, issue):
        pass

    def delete_issue(self, issue):
        pass

    def create_comment(self, comment):
        pass

    def update_comment(self, comment):
        pass

    def delete_comment(self, comment):
        pass

    def get_users(self):
        """
        This method should return all users that are related to support project on backend.

        Each user should be represented as not saved SupportUser instance.
        """
        pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from decimal import Decimal
import nodeconductor.core.fields
import django.core.validators
import nodeconductor.core.validators


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='PackageComponent',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('type', models.CharField(max_length=50, choices=[('ram', 'RAM, MB'), ('cores', 'Cores'), ('storage', 'Storage, MB')])),
                ('amount', models.PositiveIntegerField(default=0)),
                ('price', models.DecimalField(default=0, help_text='The price per unit of amount', verbose_name='Price per hour', max_digits=13, decimal_places=7, validators=[django.core.validators.MinValueValidator(Decimal('0'))])),
            ],
        ),
        migrations.CreateModel(
            name='PackageTemplate',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[nodeconductor.core.validators.validate_name])),
                ('icon_url', models.URLField(verbose_name='icon url', blank=True)),
                ('uuid', nodeconductor.core.fields.UUIDField()),
                ('type', models.CharField(default='openstack', max_length=50, choices=[('openstack', 'OpenStack')])),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='packagecomponent',
            name='template',
            field=models.ForeignKey(related_name='components', to='packages.PackageTemplate'),
        ),
        migrations.AlterUniqueTogether(
            name='packagecomponent',
            unique_together=set([('type', 'template')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-18 16:59
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0008_add_product_code'),
    ]

    operations = [
        migrations.AddField(
            model_name='offering',
            name='article_code',
            field=models.CharField(blank=True, max_length=30),
        ),
    ]



import logging

from django.db import transaction

from nodeconductor.structure import models as structure_models

from . import log, models

logger = logging.getLogger(__name__)


def create_price_estimate(sender, instance, created=False, **kwargs):
    if not created:
        return
    models.PriceEstimate.objects.create(scope=instance)


def delete_stale_price_estimate(sender, instance, **kwargs):
    models.PriceEstimate.objects.filter(scope=instance).delete()


def process_invoice_item(sender, instance, created=False, **kwargs):
    if (not created and
            not instance.tracker.has_changed('unit_price') and
            not instance.tracker.has_changed('start') and
            not instance.tracker.has_changed('end')):
        return
    with transaction.atomic():
        for scope in [instance.project, instance.project.customer]:
            estimate, _ = models.PriceEstimate.objects.get_or_create(scope=scope)
            estimate.update_total()
            estimate.validate_limit()
            estimate.save(update_fields=['total'])


def log_price_estimate_limit_update(sender, instance, created=False, **kwargs):
    if created:
        return

    if instance.tracker.has_changed('limit'):
        if isinstance(instance.scope, structure_models.Customer):
            event_type = 'project_price_limit_updated'
        elif isinstance(instance.scope, structure_models.Project):
            event_type = 'customer_price_limit_updated'
        else:
            logger.warning('A price estimate event for type of "%s" is not registered.', type(instance.scope))
            return

        message = 'Price limit for "%(scope)s" has been updated from "%(old)s" to "%(new)s".' % {
            'scope': instance.scope,
            'old': instance.tracker.previous('limit'),
            'new': instance.limit
        }
        log.event_logger.price_estimate.info(message, event_type=event_type)



from __future__ import unicode_literals

from django.apps import AppConfig
from django.db.models import signals


class ExpertsConfig(AppConfig):
    name = 'nodeconductor_assembly_waldur.experts'
    verbose_name = 'Experts'

    def ready(self):
        from nodeconductor_assembly_waldur.invoices import registrators as invoices_registrators
        from . import handlers, registrators

        ExpertRequest = self.get_model('ExpertRequest')
        ExpertBid = self.get_model('ExpertBid')

        invoices_registrators.RegistrationManager.add_registrator(
            ExpertRequest,
            registrators.ExpertRequestRegistrator
        )

        signals.post_save.connect(
            handlers.add_completed_expert_request_to_invoice,
            sender=ExpertRequest,
            dispatch_uid='nodeconductor_assembly_waldur.experts.handlers.add_completed_expert_request_to_invoice',
        )

        signals.pre_delete.connect(
            handlers.terminate_invoice_when_expert_request_deleted,
            sender=ExpertRequest,
            dispatch_uid='nodeconductor_assembly_waldur.experts.handlers.terminate_invoice_when_expert_request_deleted',
        )

        signals.post_save.connect(
            handlers.log_expert_request_creation,
            sender=ExpertRequest,
            dispatch_uid='nodeconductor_assembly_waldur.experts.handlers.log_expert_request_creation',
        )

        signals.post_save.connect(
            handlers.log_expert_request_state_changed,
            sender=ExpertRequest,
            dispatch_uid='nodeconductor_assembly_waldur.experts.handlers.log_expert_request_state_changed',
        )

        signals.post_save.connect(
            handlers.log_expert_bid_creation,
            sender=ExpertBid,
            dispatch_uid='nodeconductor_assembly_waldur.experts.handlers.log_expert_bid_creation',
        )



from django.apps import AppConfig
from django.db.models import signals


class SlurmInvoicesConfig(AppConfig):
    name = 'nodeconductor_assembly_waldur.slurm_invoices'
    verbose_name = 'Batch packages'

    def ready(self):
        from nodeconductor_assembly_waldur.invoices import registrators
        from waldur_slurm import models as slurm_models
        from . import handlers, registrators as slurm_registrators

        registrators.RegistrationManager.add_registrator(
            slurm_models.Allocation,
            slurm_registrators.AllocationRegistrator
        )

        signals.post_save.connect(
            handlers.add_new_allocation_to_invoice,
            sender=slurm_models.Allocation,
            dispatch_uid='waldur_slurm.handlers.add_new_allocation_to_invoice',
        )

        signals.post_save.connect(
            handlers.terminate_invoice_when_allocation_cancelled,
            sender=slurm_models.Allocation,
            dispatch_uid='waldur_slurm.handlers.terminate_invoice_when_allocation_cancelled',
        )

        signals.pre_delete.connect(
            handlers.terminate_invoice_when_allocation_deleted,
            sender=slurm_models.Allocation,
            dispatch_uid='waldur_slurm.handlers.terminate_invoice_when_allocation_deleted',
        )

        signals.post_save.connect(
            handlers.update_invoice_item_on_allocation_usage_update,
            sender=slurm_models.Allocation,
            dispatch_uid='waldur_slurm.handlers.update_invoice_item_on_allocation_usage_update',
        )



from rest_framework import serializers

from . import models


class SlurmPackageSerializer(serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.SlurmPackage
        fields = ('uuid', 'url', 'service_settings',  'name', 'cpu_price', 'gpu_price', 'ram_price')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'slurm-package-detail'},
            'service_settings': {'lookup_field': 'uuid', 'read_only': True},
        }



from django.conf import settings
from django.contrib import admin
from django.forms import ModelForm, ChoiceField

from nodeconductor.core import admin as core_admin
from nodeconductor.structure import admin as structure_admin

from . import models


class InvoiceItemInline(core_admin.UpdateOnlyModelAdmin, admin.TabularInline):
    model = models.InvoiceItem
    readonly_fields = ('name', 'price', 'unit_price', 'unit', 'start', 'end',
                       'project_name', 'project_uuid', 'product_code', 'article_code')
    exclude = ('project',)


class OpenStackItemInline(InvoiceItemInline):
    model = models.OpenStackItem
    readonly_fields = InvoiceItemInline.readonly_fields + ('package', 'package_details')


class OfferingItemInline(InvoiceItemInline):
    model = models.OfferingItem
    readonly_fields = InvoiceItemInline.readonly_fields + ('offering', 'offering_details')


class InvoiceAdmin(core_admin.UpdateOnlyModelAdmin, admin.ModelAdmin):
    inlines = [OpenStackItemInline, OfferingItemInline]
    readonly_fields = ('customer', 'state', 'total', 'year', 'month')
    list_display = ('customer', 'total', 'year', 'month', 'state')
    list_filter = ('state', 'customer')
    search_fields = ('customer', 'uuid')


class PaymentDetailsInline(admin.StackedInline):
    model = models.PaymentDetails

    def get_readonly_fields(self, request, obj=None):
        fields = super(PaymentDetailsInline, self).get_readonly_fields(request, obj)
        if obj and hasattr(obj, 'payment_details') and obj.payment_details.is_billable():
            fields += ('accounting_start_date',)
        return fields


class PaymentDetailsAdminForm(ModelForm):
    def __init__(self, *args, **kwargs):
        super(PaymentDetailsAdminForm, self).__init__(*args, **kwargs)
        self.fields['type'] = ChoiceField(choices=[(t, t) for t in settings.INVOICES['COMPANY_TYPES']])


class PaymentDetailsAdmin(admin.ModelAdmin):
    form = PaymentDetailsAdminForm

    def get_readonly_fields(self, request, obj=None):
        fields = super(PaymentDetailsAdmin, self).get_readonly_fields(request, obj)
        if obj and obj.is_billable():
            fields += ('accounting_start_date',)
        return fields


structure_admin.CustomerAdmin.inlines += [PaymentDetailsInline]
admin.site.register(models.Invoice, InvoiceAdmin)
admin.site.register(models.PaymentDetails, PaymentDetailsAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import model_utils.fields
import nodeconductor.core.fields
import django.utils.timezone
from django.conf import settings
import nodeconductor.core.validators


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('support', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Comment',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, verbose_name='created', editable=False)),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, verbose_name='modified', editable=False)),
                ('uuid', nodeconductor.core.fields.UUIDField()),
                ('description', models.TextField()),
                ('is_public', models.BooleanField(default=True)),
                ('backend_id', models.CharField(max_length=255, blank=True)),
            ],
            options={
                'ordering': ['-created'],
            },
        ),
        migrations.CreateModel(
            name='SupportUser',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[nodeconductor.core.validators.validate_name])),
                ('uuid', nodeconductor.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, blank=True)),
                ('user', models.ForeignKey(related_name='+', blank=True, to=settings.AUTH_USER_MODEL, null=True)),
            ],
            options={
                'abstract': False,
                'ordering': ['name'],
            },
        ),
        migrations.RemoveField(
            model_name='issue',
            name='creator',
        ),
        migrations.AddField(
            model_name='issue',
            name='backend_id',
            field=models.CharField(max_length=255, blank=True),
        ),
        migrations.AddField(
            model_name='issue',
            name='deadline',
            field=models.DateTimeField(null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='issue',
            name='assignee',
            field=models.ForeignKey(related_name='issues', blank=True, to='support.SupportUser', null=True),
        ),
        migrations.AlterField(
            model_name='issue',
            name='key',
            field=models.CharField(max_length=255, blank=True),
        ),
        migrations.AlterField(
            model_name='issue',
            name='reporter',
            field=models.ForeignKey(related_name='reported_issues', to='support.SupportUser'),
        ),
        migrations.AddField(
            model_name='comment',
            name='author',
            field=models.ForeignKey(related_name='comments', to='support.SupportUser'),
        ),
        migrations.AddField(
            model_name='comment',
            name='issue',
            field=models.ForeignKey(related_name='comments', to='support.Issue'),
        ),
        migrations.AddField(
            model_name='issue',
            name='caller',
            field=models.ForeignKey(related_name='created_issues', to='support.SupportUser'),
        ),
    ]



from __future__ import unicode_literals

import copy

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import transaction
from django.template import Context, Template
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from nodeconductor.core import serializers as core_serializers, utils as core_utils
from nodeconductor.structure import models as structure_models, SupportedServices, serializers as structure_serializers
from nodeconductor_assembly_waldur.support.backend.atlassian import ServiceDeskBackend

from . import models, backend

User = get_user_model()


def render_issue_template(config_name, issue):
    issue_settings = settings.WALDUR_SUPPORT.get('ISSUE', {})
    if not issue_settings:
        return ''

    raw = issue_settings[config_name]
    template = Template(raw)
    return template.render(Context({'issue': issue}))


class IssueSerializer(core_serializers.AugmentedSerializerMixin,
                      serializers.HyperlinkedModelSerializer):
    resource = core_serializers.GenericRelatedField(
        related_models=structure_models.ResourceMixin.get_all_models(), required=False)
    caller = serializers.HyperlinkedRelatedField(
        view_name='user-detail',
        lookup_field='uuid',
        queryset=User.objects.all(),
        required=False,
        allow_null=True,
    )
    reporter = serializers.HyperlinkedRelatedField(
        view_name='support-user-detail',
        lookup_field='uuid',
        read_only=True
    )
    assignee = serializers.HyperlinkedRelatedField(
        view_name='support-user-detail',
        lookup_field='uuid',
        queryset=models.SupportUser.objects.all(),
        required=False,
        allow_null=True,
    )
    resource_type = serializers.SerializerMethodField()
    resource_name = serializers.ReadOnlyField(source='resource.name')
    type = serializers.ChoiceField(
        choices=[(t, t) for t in settings.WALDUR_SUPPORT['ISSUE']['types']],
        initial=settings.WALDUR_SUPPORT['ISSUE']['types'][0],
        default=settings.WALDUR_SUPPORT['ISSUE']['types'][0])
    is_reported_manually = serializers.BooleanField(
        initial=False, default=False, write_only=True,
        help_text=_('Set true if issue is created by regular user via portal.'))

    class Meta(object):
        model = models.Issue
        fields = (
            'url', 'uuid', 'type', 'key', 'backend_id', 'link',
            'summary', 'description', 'status', 'resolution', 'priority',
            'caller', 'caller_uuid', 'caller_full_name',
            'reporter', 'reporter_uuid', 'reporter_name',
            'assignee', 'assignee_uuid', 'assignee_name',
            'customer', 'customer_uuid', 'customer_name',
            'project', 'project_uuid', 'project_name',
            'resource', 'resource_type', 'resource_name',
            'created', 'modified', 'is_reported_manually',
            'first_response_sla',
        )
        read_only_fields = ('key', 'status', 'resolution', 'backend_id', 'link', 'priority', 'first_response_sla')
        protected_fields = ('customer', 'project', 'resource', 'type', 'caller')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            customer={'lookup_field': 'uuid', 'view_name': 'customer-detail'},
            project={'lookup_field': 'uuid', 'view_name': 'project-detail'},
        )
        related_paths = dict(
            caller=('uuid', 'full_name',),
            reporter=('uuid', 'name',),
            assignee=('uuid', 'name',),
            customer=('uuid', 'name',),
            project=('uuid', 'name',),
        )

    def get_fields(self):
        fields = super(IssueSerializer, self).get_fields()

        if 'view' not in self.context:  # On docs generation context does not contain "view".
            return fields

        user = self.context['view'].request.user
        if not user.is_staff and not user.is_support:
            del fields['link']

        return fields

    def get_resource_type(self, obj):
        if obj.resource:
            return SupportedServices.get_name_for_model(obj.resource_content_type.model_class())

    def validate(self, attrs):
        if self.instance is not None:
            return attrs
        if attrs.pop('is_reported_manually'):
            attrs['caller'] = self.context['request'].user
            if attrs.get('assignee'):
                raise serializers.ValidationError(
                    {'assignee': _('Assignee cannot be defined if issue is reported manually.')})
        else:
            if not attrs.get('caller'):
                raise serializers.ValidationError({'caller': _('This field is required.')})
            reporter = models.SupportUser.objects.filter(user=self.context['request'].user).first()
            if not reporter:
                raise serializers.ValidationError(
                    _('You cannot report issues because your help desk account is not connected to profile.'))
            attrs['reporter'] = reporter
        return attrs

    def validate_customer(self, customer):
        """ User has to be customer owner, staff or global support """
        if not customer:
            return customer
        user = self.context['request'].user
        if (not customer or
                user.is_staff or
                user.is_support or
                customer.has_user(user, structure_models.CustomerRole.OWNER)):
            return customer
        raise serializers.ValidationError(_('Only customer owner, staff or support can report customer issues.'))

    def validate_project(self, project):
        if not project:
            return project
        user = self.context['request'].user
        if (not project or
                user.is_staff or
                user.is_support or
                project.customer.has_user(user, structure_models.CustomerRole.OWNER) or
                project.has_user(user, structure_models.ProjectRole.MANAGER) or
                project.has_user(user, structure_models.ProjectRole.ADMINISTRATOR)):
            return project
        raise serializers.ValidationError(
            _('Only customer owner, project manager, project admin, staff or support can report such issue.'))

    def validate_resource(self, resource):
        if resource:
            self.validate_project(resource.service_project_link.project)
        return resource

    @transaction.atomic()
    def create(self, validated_data):
        resource = validated_data.get('resource')
        if resource:
            validated_data['project'] = resource.service_project_link.project
        project = validated_data.get('project')
        if project:
            validated_data['customer'] = project.customer

        validated_data['description'] = render_issue_template('description', validated_data)
        validated_data['summary'] = render_issue_template('summary', validated_data)
        return super(IssueSerializer, self).create(validated_data)

    def _render_template(self, config_name, issue):
        raw = self.issue_settings[config_name]
        template = Template(raw)
        return template.render(Context({'issue': issue}))


class CommentSerializer(core_serializers.AugmentedSerializerMixin,
                        serializers.HyperlinkedModelSerializer):
    # should be initialized with issue in context on creation
    author_user = serializers.HyperlinkedRelatedField(
        source='author.user',
        view_name='user-detail',
        lookup_field='uuid',
        read_only=True,
    )

    class Meta(object):
        model = models.Comment
        fields = ('url', 'uuid', 'issue', 'issue_key', 'description', 'is_public',
                  'author_name', 'author_user', 'backend_id', 'created')
        read_only_fields = ('issue', 'backend_id',)
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
        )
        related_paths = dict(
            author=('name',),
            issue=('key',),
        )

    @transaction.atomic()
    def create(self, validated_data):
        author_user = self.context['request'].user
        validated_data['author'], _ = models.SupportUser.objects.get_or_create_from_user(author_user)
        validated_data['issue'] = self.context['view'].get_object()
        return super(CommentSerializer, self).create(validated_data)


class SupportUserSerializer(core_serializers.AugmentedSerializerMixin,
                            serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.SupportUser
        fields = ('url', 'uuid', 'name', 'backend_id', 'user')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid'},
            user={'lookup_field': 'uuid', 'view_name': 'user-detail'}
        )


class WebHookReceiverSerializer(serializers.Serializer):
    class EventType:
        CREATED = 'jira:issue_created'
        UPDATED = 'jira:issue_updated'
        DELETED = 'jira:issue_deleted'

    PUBLIC_COMMENT_KEY = 'sd.public.comment'

    def validate(self, attrs):
        if 'issue' not in self.initial_data:
            raise serializers.ValidationError(_('"issue" is missing in request data. Cannot process issue.'))

        if 'webhookEvent' not in self.initial_data:
            raise serializers.ValidationError(_('"webhookEvent" is missing in request data. Cannot find out even type'))

        return attrs

    @transaction.atomic()
    def save(self, **kwargs):
        fields = self.initial_data['issue']['fields']
        backend_id = self.initial_data['issue']['key']

        event_type = self.initial_data['webhookEvent']
        if event_type == self.EventType.UPDATED:
            try:
                issue = models.Issue.objects.get(backend_id=backend_id)
            except models.Issue.DoesNotExist:
                pass
            else:
                backend_issue = self._get_backend_issue(fields=fields)
                self._update_issue(issue=issue, backend_issue=backend_issue)

                if 'comment' in fields:
                    backend_comments = self._get_backend_comments(issue.key, fields)
                    self._update_comments(issue=issue, backend_comments=backend_comments)

        elif event_type == self.EventType.DELETED:
            issue = models.Issue.objects.get(backend_id=backend_id)
            issue.delete()

    def _get_backend_issue(self, fields):
        """
        Builds a dictionary of issue attributes and values read from a JIRA response.
        :param fields: issue fields in a response;
        :return: a dictionary of issue attributes and values
        """
        backend_issue = {
            'resolution': self._get_field_name(fields, 'resolution'),
            'status': self._get_field_name(fields, 'status'),
            'impact': self._get_impact_field(fields=fields),
            'summary': fields['summary'],
            'priority': self._get_field_name(fields, 'priority'),
            'description': fields['description'] or '',
            'type': fields['issuetype']['name'],
        }

        custom_field_values = [fields[customfield] for customfield in fields if customfield.startswith('customfield')]
        backend_issue['first_response_sla'] = self._get_sla_field_value(custom_field_values)

        assignee = self._get_support_user_by_field_name(field_name='assignee', fields=fields)
        if assignee:
            backend_issue['assignee'] = assignee

        reporter = self._get_support_user_by_field_name(field_name='reporter', fields=fields)
        if reporter:
            backend_issue['reporter'] = reporter
            if reporter.user:
                backend_issue['caller'] = reporter.user

        return backend_issue

    def _update_issue(self, issue, backend_issue):
        """
        Updates given issue from the backend issue if it has been changed.
        :param issue: an issue to update
        :param backend_issue: a set of parameters to be updated.
        """
        updated = False

        for field, backend_value in backend_issue.items():
            current_value = getattr(issue, field)
            if current_value != backend_value:
                setattr(issue, field, backend_value)
                updated = True

        if updated:
            issue.save()

    def _get_field_name(self, fields, field_name, default_value=''):
        """ Returns 'name' attribute of the field or default_value if the value is None """
        return default_value if not fields[field_name] else fields[field_name]['name']

    def _get_sla_field_value(self, custom_field_values):
        sla_field_name = settings.WALDUR_SUPPORT.get('ISSUE', {}).get('sla_field', None)

        for field in custom_field_values:
            if isinstance(field, dict):
                name = field.get('name', None)
                if name and name == sla_field_name:
                    ongoing_cycle = field.get('ongoingCycle', {})
                    breach_time = ongoing_cycle.get('breachTime', {})
                    epoch_milliseconds = breach_time.get('epochMillis', None)
                    if epoch_milliseconds:
                        return core_utils.timestamp_to_datetime(epoch_milliseconds / 1000.0)

        return None

    def _get_impact_field(self, fields):
        issue_settings = settings.WALDUR_SUPPORT.get('ISSUE', {})
        impact_field_name = issue_settings.get('impact_field', None)
        return fields.get(impact_field_name, '')

    def _get_backend_comments(self, issue_key, fields):
        """
        Forms a dictionary of a backend comments where an id is a key and a comment body is a value.
        :param issue_key: an issue key to look up for comments;
        :param fields: fields from issue in the response;
        :return: a dictionary of a backend comments with their ids as keys.
        """
        active_backend = backend.get_active_backend()
        if self._is_service_desk():
            comments = active_backend.expand_comments(issue_key)
            backend_comments = {c['id']: c for c in comments}
        else:
            backend_comments = {c['id']: c for c in fields['comment']['comments']}

        return backend_comments

    def _is_service_desk(self):
        """
        :return: True if current JIRA instance is ServiceDesk, otherwise False.
        """
        return isinstance(backend.get_active_backend(), ServiceDeskBackend)

    @transaction.atomic()
    def _update_comments(self, issue, backend_comments):
        comments = {c.backend_id: c for c in issue.comments.all()}
        active_backend = backend.get_active_backend()

        for exist_comment_id in set(backend_comments) & set(comments):
            backend_comment = backend_comments[exist_comment_id]
            comment = comments[exist_comment_id]

            update_fields = []

            backend_comment_description = active_backend.extract_comment_message(backend_comment['body'])
            if comment.description != backend_comment_description:
                comment.description = backend_comment_description
                update_fields.append('description')

            if self._is_service_desk():
                is_public = self._get_comment_public_field_value(backend_comment)
                if is_public != comment.is_public:
                    comment.is_public = is_public
                    update_fields.append('is_public')

            if update_fields:
                comment.save(update_fields=update_fields)

        for new_comment_id in set(backend_comments) - set(comments):
            backend_comment = backend_comments[new_comment_id]
            author, _ = models.SupportUser.objects.get_or_create(backend_id=backend_comment['author']['key'])
            new_comment = models.Comment(
                issue=issue,
                author=author,
                description=backend_comment['body'],
                backend_id=backend_comment['id'],
            )

            if self._is_service_desk():
                new_comment.is_public = self._get_comment_public_field_value(backend_comment)

            new_comment.save()

        models.Comment.objects.filter(backend_id__in=set(comments) - set(backend_comments)).delete()

    def _get_comment_public_field_value(self, backend_comment):
        properties = backend_comment.get('properties', {})

        try:
            internal_property = next(p for p in properties if p.get('key') == self.PUBLIC_COMMENT_KEY)
        except StopIteration:
            return True

        is_internal = internal_property.get('value', {}).get('internal', False)

        return not is_internal

    def _get_support_user_by_field_name(self, fields, field_name):
        support_user = None

        if field_name in fields and fields[field_name]:
            support_user_backend_key = fields[field_name]['key']

            if support_user_backend_key:
                support_user, _ = models.SupportUser.objects.get_or_create(backend_id=support_user_backend_key)

        return support_user


class OfferingSerializer(structure_serializers.PermissionFieldFilteringMixin,
                         core_serializers.AugmentedSerializerMixin,
                         serializers.HyperlinkedModelSerializer):
    type = serializers.ChoiceField(choices=settings.WALDUR_SUPPORT['OFFERINGS'].keys())
    state = serializers.ReadOnlyField(source='get_state_display')

    class Meta(object):
        model = models.Offering
        fields = ('url', 'uuid', 'name', 'project', 'type', 'state', 'type_label', 'unit_price',
                  'unit', 'created', 'modified', 'issue', 'issue_name', 'issue_link',
                  'issue_key', 'issue_description', 'issue_uuid', 'issue_status',
                  'project_name', 'project_uuid', 'product_code', 'article_code')
        read_only_fields = ('type_label', 'issue', 'unit_price', 'unit', 'state', 'product_code', 'article_code')
        protected_fields = ('project', 'type')
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'support-offering-detail'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
            project={'lookup_field': 'uuid', 'view_name': 'project-detail'},
        )
        related_paths = dict(
            issue=('uuid', 'name', 'status', 'key', 'description', 'link'),
            project=('uuid', 'name',),
        )

    def get_filtered_field_names(self):
        return ('project',)


class ConfigurableSerializerMixin(object):
    """
    Serializer is built on top WALDUR_SUPPORT['OFFERINGS'] configuration.

    Each configured field get's converted to serializer field according to field type in the configuration.

    Field types:
        'integer' - corresponds to 'serializers.IntegerField'
        'string' - is a default field type even if it is not defined explicitly in configuration.
                   Corresponds to 'serializers.CharField(max_length=255)'

    Default values:
        if 'default' key is present in option field configuration it is going to be used in serializer unless
        the value itself has been provided in the create request.
    """

    def _get_offerings_configuration(self):
        return copy.deepcopy(settings.WALDUR_SUPPORT['OFFERINGS'])

    def _get_configuration(self, type):
        return self._get_offerings_configuration().get(type)

    def get_fields(self):
        result = super(ConfigurableSerializerMixin, self).get_fields()
        if hasattr(self, 'initial_data') and not hasattr(self, '_errors'):
            type = self.initial_data['type']
            configuration = self._get_configuration(type)
            for attr_name in configuration['order']:
                attr_options = configuration['options'].get(attr_name, {})
                result[attr_name] = self._get_field_instance(attr_options)

        # choices have to be added dynamically so that unit tests can mock offering configuration.
        # otherwise it is always going to be a default set up.
        result['type'] = serializers.ChoiceField(allow_blank=False, choices=self._get_offerings_configuration().keys())

        return result

    def _validate_type(self, type):
        offering_configuration = self._get_configuration(type)
        if offering_configuration is None:
            raise serializers.ValidationError({
                'type': _('Type configuration could not be found.')
            })

    def validate_empty_values(self, data):
        if 'type' not in data or ('type' in data and data['type'] is None):
            raise serializers.ValidationError({
                'type': _('This field is required.')
            })
        else:
            self._validate_type(data['type'])

        return super(ConfigurableSerializerMixin, self).validate_empty_values(data)

    def _get_field_instance(self, attr_options):
        field_type = attr_options.get('type', '').lower()

        if field_type == 'string':
            field = serializers.CharField(max_length=255, write_only=True)
        elif field_type == 'integer':
            field = serializers.IntegerField(write_only=True)
        else:
            field = serializers.CharField(write_only=True)

        default_value = attr_options.get('default')
        if default_value:
            field.default = default_value

        field.required = attr_options.get('required', False)
        field.label = attr_options.get('label')
        field.help_text = attr_options.get('help_text')

        return field

    def _form_description(self, configuration, validated_data):
        result = []

        for key in configuration['order']:
            if key not in validated_data:
                continue

            label = configuration['options'].get(key, {})
            label_value = label.get('label', key)
            result.append('%s: \'%s\'' % (label_value, validated_data[key]))

        if 'description' in validated_data:
            result.append('\n %s' % validated_data['description'])

        return '\n'.join(result)

    def _get_extra(self, configuration, validated_data):
        result = {
            key: validated_data[key]
            for key in configuration['order']
            if key in validated_data
        }
        if 'description' in validated_data:
            result['description'] = validated_data['description']
        return result


class OfferingCreateSerializer(ConfigurableSerializerMixin, OfferingSerializer):
    description = serializers.CharField(required=False, help_text=_('Description to add to the issue.'))

    class Meta(OfferingSerializer.Meta):
        fields = OfferingSerializer.Meta.fields + ('description',)
        extra_kwargs = dict(
            url={'lookup_field': 'uuid', 'view_name': 'support-offering-detail'},
            issue={'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
            project={'lookup_field': 'uuid',
                     'view_name': 'project-detail',
                     'required': True,
                     'allow_empty': False,
                     'allow_null': False},
        )

    def create(self, validated_data):
        """
        Each offering corresponds to the single issue which has next values:
            'project' - a hyperlinked field which must be provided with every request;
            'customer' - customer is extracted from the provided project;
            'caller' - a user who sent a request is considered to be a 'caller' of the issue;
            'summary' - has a format of 'Request for "OFFERING[name][label]' or 'Request for "Support" if empty;
            'description' - combined list of all other fields provided with the request;
        """
        project = validated_data['project']
        type = validated_data['type']
        offering_configuration = self._get_configuration(type)
        type_label = offering_configuration.get('label', type)
        issue_details = dict(
            caller=self.context['request'].user,
            project=project,
            customer=project.customer,
            type=settings.WALDUR_SUPPORT['DEFAULT_OFFERING_ISSUE_TYPE'],
            summary='Request for \'%s\'' % type_label,
            description=self._form_description(offering_configuration, validated_data))
        issue_details['summary'] = render_issue_template('summary', issue_details)
        issue_details['description'] = render_issue_template('description', issue_details)
        issue = models.Issue.objects.create(**issue_details)

        offering = models.Offering.objects.create(
            issue=issue,
            project=issue.project,
            name=validated_data.get('name'),
            type=type,
            product_code=offering_configuration.get('product_code', ''),
            article_code=offering_configuration.get('article_code', ''),
        )

        return offering


class OfferingCompleteSerializer(serializers.Serializer):
    unit_price = serializers.DecimalField(max_digits=13, decimal_places=7)
    unit = serializers.ChoiceField(choices=models.Offering.Units.CHOICES, default=models.Offering.Units.PER_DAY)

    def update(self, instance, validated_data):
        instance.unit_price = validated_data['unit_price']
        instance.unit = validated_data['unit']
        instance.state = models.Offering.States.OK
        instance.save(update_fields=['state', 'unit_price', 'unit'])
        return instance



import cStringIO
import logging

from celery import shared_task
from django.conf import settings
from django.core.mail import send_mail
from django.db.models import Q
from django.template.loader import render_to_string
from django.utils import timezone

from nodeconductor.core import utils as core_utils
from nodeconductor.core.csv import UnicodeDictWriter
from nodeconductor.structure import models as structure_models

from . import models, registrators, serializers


logger = logging.getLogger(__name__)


@shared_task(name='invoices.create_monthly_invoices')
def create_monthly_invoices():
    """
    - For every customer change state of the invoices for previous months from "pending" to "billed"
      and freeze their items.
    - Create new invoice for every customer in current month if not created yet.
    """
    date = timezone.now()

    old_invoices = models.Invoice.objects.filter(
        Q(state=models.Invoice.States.PENDING, year__lt=date.year) |
        Q(state=models.Invoice.States.PENDING, year=date.year, month__lt=date.month)
    )
    for invoice in old_invoices:
        invoice.set_created()
        invoice.freeze()

    customers = structure_models.Customer.objects.all()
    if settings.INVOICES['ENABLE_ACCOUNTING_START_DATE']:
        customers = customers.filter(
            Q(payment_details__accounting_start_date__lt=timezone.now()) |
            Q(payment_details__isnull=True)
        )
    for customer in customers.iterator():
        registrators.RegistrationManager.get_or_create_invoice(customer, core_utils.month_start(date))


@shared_task(name='invoices.send_invoice_notification')
def send_invoice_notification(invoice_uuid, link_template):
    """ Sends email notification with invoice link to customer owners """
    invoice = models.Invoice.objects.get(uuid=invoice_uuid)

    context = {
        'month': invoice.month,
        'year': invoice.year,
        'customer': invoice.customer.name,
        'link': link_template.format(uuid=invoice_uuid)
    }

    subject = render_to_string('invoices/notification_subject.txt', context)
    text_message = render_to_string('invoices/notification_message.txt', context)
    html_message = render_to_string('invoices/notification_message.html', context)

    emails = [owner.email for owner in invoice.customer.get_owners()]

    logger.debug('About to send invoice {invoice} notification to {emails}'.format(invoice=invoice, emails=emails))
    send_mail(subject, text_message, settings.DEFAULT_FROM_EMAIL, emails, html_message=html_message)


@shared_task(name='invoices.send_invoice_report')
def send_invoice_report(invoice_uuid):
    """ Sends accounting data as CSV """
    invoice = models.Invoice.objects.get(uuid=invoice_uuid)

    context = {
        'month': invoice.month,
        'year': invoice.year,
        'customer': invoice.customer.name,
    }

    subject = render_to_string('invoices/report_subject.txt', context)
    text_message = format_invoice_csv(invoice)
    emails = [settings.INVOICES['INVOICE_REPORTING']['EMAIL']]

    logger.debug('About to send invoice {invoice} report to {emails}'.format(invoice=invoice, emails=emails))
    send_mail(subject, text_message, settings.DEFAULT_FROM_EMAIL, emails)


def format_invoice_csv(invoice):
    csv_params = settings.INVOICES['INVOICE_REPORTING']['CSV_PARAMS']

    if settings.INVOICES['INVOICE_REPORTING'].get('USE_SAF'):
        fields = serializers.SAFReportSerializer.Meta.fields
        stream = cStringIO.StringIO()
        writer = UnicodeDictWriter(stream, fieldnames=fields, **csv_params)
        writer.writeheader()

        serializer = serializers.SAFReportSerializer(invoice.items, many=True)
        writer.writerows(serializer.data)
        return stream.getvalue().decode('utf-8')

    fields = serializers.InvoiceItemReportSerializer.Meta.fields
    stream = cStringIO.StringIO()
    writer = UnicodeDictWriter(stream, fieldnames=fields, **csv_params)
    writer.writeheader()

    openstack_items = invoice.openstack_items.all().select_related('invoice__customer')
    openstack_serializer = serializers.OpenStackItemReportSerializer(openstack_items, many=True)
    writer.writerows(openstack_serializer.data)

    offering_items = invoice.offering_items.all().select_related('invoice__customer')
    offering_serializer = serializers.OfferingItemReportSerializer(offering_items, many=True)
    writer.writerows(offering_serializer.data)

    generic_items = invoice.generic_items.all().select_related('invoice__customer')
    generic_serializer = serializers.GenericItemReportSerializer(generic_items, many=True)
    writer.writerows(generic_serializer.data)

    return stream.getvalue().decode('utf-8')



from __future__ import unicode_literals

from nodeconductor.core import filters as core_filters
from nodeconductor.cost_tracking import models


class PriceEstimateScopeFilterBackend(core_filters.GenericKeyFilterBackend):

    def get_related_models(self):
        return models.PriceEstimate.get_estimated_models()

    def get_field_name(self):
        return 'scope'



from __future__ import unicode_literals

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker

from model_utils.models import TimeStampedModel

from nodeconductor.core import models as core_models
from nodeconductor.structure import models as structure_models
from nodeconductor_assembly_waldur.common import mixins as common_mixins

from . import managers


@python_2_unicode_compatible
class Issue(core_models.UuidMixin,
            structure_models.StructureLoggableMixin,
            core_models.BackendModelMixin,
            TimeStampedModel):
    class Meta:
        ordering = ['-created']

    class Permissions(object):
        customer_path = 'customer'
        project_path = 'project'

    backend_id = models.CharField(max_length=255, blank=True)
    key = models.CharField(max_length=255, blank=True)
    type = models.CharField(max_length=255)
    link = models.URLField(max_length=255, help_text=_('Link to issue in support system.'), blank=True)

    summary = models.CharField(max_length=255)
    description = models.TextField(blank=True)
    deadline = models.DateTimeField(blank=True, null=True)
    impact = models.CharField(max_length=255, blank=True)

    status = models.CharField(max_length=255)
    resolution = models.CharField(max_length=255, blank=True)
    priority = models.CharField(max_length=255, blank=True)

    caller = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='created_issues',
                               help_text=_('Waldur user who has reported the issue.'))
    reporter = models.ForeignKey('SupportUser', related_name='reported_issues', blank=True, null=True,
                                 help_text=_('Help desk user who have created the issue that is reported by caller.'))
    assignee = models.ForeignKey('SupportUser', related_name='issues', blank=True, null=True,
                                 help_text=_('Help desk user who will implement the issue'))

    customer = models.ForeignKey(
        structure_models.Customer, verbose_name=_('organization'), related_name='issues', blank=True, null=True)
    project = models.ForeignKey(structure_models.Project, related_name='issues', blank=True, null=True)

    resource_content_type = models.ForeignKey(ContentType, null=True)
    resource_object_id = models.PositiveIntegerField(null=True)
    resource = GenericForeignKey('resource_content_type', 'resource_object_id')

    first_response_sla = models.DateTimeField(blank=True, null=True)

    tracker = FieldTracker()

    @classmethod
    def get_url_name(cls):
        return 'support-issue'

    @classmethod
    def get_backend_fields(cls):
        return super(Issue, cls).get_backend_fields() + ('backend_id', 'key', 'type', 'link',
                                                         'summary', 'description', 'deadline', 'impact',
                                                         'status', 'resolution', 'priority',
                                                         'caller', 'reporter', 'assignee', 'customer', 'project',
                                                         'resource', 'first_response_sla')

    def get_log_fields(self):
        return ('uuid', 'type', 'key', 'status', 'link', 'summary',
                'reporter', 'caller', 'customer', 'project', 'resource')

    def __str__(self):
        return '{}: {}'.format(self.key or '???', self.summary)


@python_2_unicode_compatible
class SupportUser(core_models.UuidMixin, core_models.NameMixin, models.Model):
    class Meta:
        ordering = ['name']

    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='+', blank=True, null=True)
    backend_id = models.CharField(max_length=255, blank=True)

    objects = managers.SupportUserManager()

    @classmethod
    def get_url_name(cls):
        return 'support-user'

    def __str__(self):
        return self.name


@python_2_unicode_compatible
class Comment(core_models.UuidMixin, core_models.BackendModelMixin, TimeStampedModel):
    class Meta:
        ordering = ['-created']

    class Permissions(object):
        customer_path = 'issue__customer'
        project_path = 'issue__project'

    issue = models.ForeignKey(Issue, related_name='comments')
    author = models.ForeignKey(SupportUser, related_name='comments')
    description = models.TextField()
    is_public = models.BooleanField(default=True)
    backend_id = models.CharField(max_length=255, blank=True)

    @classmethod
    def get_url_name(cls):
        return 'support-comment'

    @classmethod
    def get_backend_fields(cls):
        return super(Comment, cls).get_backend_fields() + ('issue', 'author', 'description', 'is_public', 'backend_id')

    def __str__(self):
        return self.description[:50]


@python_2_unicode_compatible
class Offering(core_models.UuidMixin,
               core_models.NameMixin,
               common_mixins.ProductCodeMixin,
               common_mixins.UnitPriceMixin,
               structure_models.StructureLoggableMixin,
               TimeStampedModel):

    class Meta:
        ordering = ['-created']

    class Permissions(object):
        customer_path = 'project__customer'
        project_path = 'project'

    class States(object):
        REQUESTED = 'requested'
        OK = 'ok'
        TERMINATED = 'terminated'

        CHOICES = ((REQUESTED, _('Requested')), (OK, _('OK')), (TERMINATED, _('Terminated')))

    type = models.CharField(max_length=255)
    issue = models.ForeignKey(Issue, null=True, on_delete=models.SET_NULL)
    project = models.ForeignKey(structure_models.Project, null=True, on_delete=models.PROTECT)
    state = models.CharField(default=States.REQUESTED, max_length=30, choices=States.CHOICES)

    tracker = FieldTracker()

    def get_log_fields(self):
        return super(Offering, self).get_log_fields() + ('state', )

    @property
    def type_label(self):
        offerings = settings.WALDUR_SUPPORT.get('OFFERINGS', {})
        type_settings = offerings.get(self.type, {})
        return type_settings.get('label', None)

    @classmethod
    def get_url_name(cls):
        return 'support-offering'

    def __str__(self):
        return '{}: {}'.format(self.type_label or self.name, self.state)

    @classmethod
    def get_scope_type(cls):
        return 'Support.Offering'

    def _get_log_context(self, entity_name):
        context = super(Offering, self)._get_log_context(entity_name)
        context['resource_type'] = self.get_scope_type()
        context['resource_uuid'] = self.uuid.hex
        return context



from django.utils import timezone

from nodeconductor_assembly_waldur.invoices import registrators as invoices_registrators

from .log import event_logger
from . import models


def log_expert_request_creation(sender, instance, created=False, **kwargs):
    if not created:
        return

    event_logger.waldur_expert_request.info(
        'User {user_username} with full name {user_full_name} has created '
        'request for experts under {customer_name} / {project_name}.',
        event_type='expert_request_created',
        event_context={
            'expert_request': instance,
        })


def log_expert_request_state_changed(sender, instance, created=False, **kwargs):
    if created:
        return

    if not instance.tracker.has_changed('state'):
        return

    if instance.state == models.ExpertRequest.States.ACTIVE:
        event_logger.waldur_expert_request.info(
            'Expert request {expert_request_name} has been activated.',
            event_type='expert_request_activated',
            event_context={
                'expert_request': instance,
            })
    elif instance.state == models.ExpertRequest.States.CANCELLED:
        event_logger.waldur_expert_request.info(
            'Expert request {expert_request_name} has been cancelled.',
            event_type='expert_request_cancelled',
            event_context={
                'expert_request': instance,
            })
    elif instance.state == models.ExpertRequest.States.COMPLETED:
        event_logger.waldur_expert_request.info(
            'Expert request {expert_request_name} has been completed.',
            event_type='expert_request_completed',
            event_context={
                'expert_request': instance,
            })


def log_expert_bid_creation(sender, instance, created=False, **kwargs):
    if not created:
        return

    event_logger.waldur_expert_bid.info(
        'User {user_username} with full name {user_full_name} has created '
        'bid for request {request_name} under {customer_name} / {project_name}.',
        event_type='expert_bid_created',
        event_context={
            'expert_bid': instance,
        })


def add_completed_expert_request_to_invoice(sender, instance, created=False, **kwargs):
    if created:
        return

    state = instance.state
    if state != models.ExpertRequest.States.COMPLETED or state == instance.tracker.previous('state'):
        return

    if not instance.issue or not hasattr(instance, 'contract'):
        return

    invoices_registrators.RegistrationManager.register(instance, timezone.now())


def terminate_invoice_when_expert_request_deleted(sender, instance, **kwargs):
    invoices_registrators.RegistrationManager.terminate(instance, timezone.now())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('packages', '0009_set_tenant_extra_configuration'),
    ]

    operations = [
        migrations.AddField(
            model_name='packagetemplate',
            name='archived',
            field=models.BooleanField(default=False, help_text='Forbids creation of new packages.'),
        ),
        migrations.AlterField(
            model_name='openstackpackage',
            name='template',
            field=models.ForeignKey(related_name='openstack_packages', on_delete=django.db.models.deletion.PROTECT, to='packages.PackageTemplate', help_text='Tenant will be created based on this template.'),
        ),
    ]



from __future__ import unicode_literals

from django.contrib import admin

from . import models


class ExpertProviderAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'customer', 'created')


class ExpertRequestAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'project', 'state', 'created')
    readonly_fields = ('project', 'created')


admin.site.register(models.ExpertProvider, ExpertProviderAdmin)
admin.site.register(models.ExpertRequest, ExpertRequestAdmin)



from __future__ import unicode_literals

from django.conf import settings
from django.db.models import Q
from django.utils import timezone

from nodeconductor.cost_tracking import signals as cost_signals
from nodeconductor_assembly_waldur.support import models as support_models

from . import tasks, models, log, registrators


def add_new_openstack_package_details_to_invoice(sender, instance, created=False, **kwargs):
    if not created:
        return

    registrators.RegistrationManager.register(instance, timezone.now())


def update_invoice_on_openstack_package_deletion(sender, instance, **kwargs):
    registrators.RegistrationManager.terminate(instance, timezone.now())


def update_invoice_on_offering_deletion(sender, instance, **kwargs):
    state = instance.state
    if state == support_models.Offering.States.TERMINATED:
        # no need to terminate offering item if it was already terminated before.
        return

    registrators.RegistrationManager.terminate(instance, timezone.now())


def add_new_offering_details_to_invoice(sender, instance, created=False, **kwargs):
    state = instance.state
    if (state == support_models.Offering.States.OK
            and support_models.Offering.States.REQUESTED == instance.tracker.previous('state')):
        registrators.RegistrationManager.register(instance, timezone.now())
    if (state == support_models.Offering.States.TERMINATED
            and support_models.Offering.States.OK == instance.tracker.previous('state')):
        registrators.RegistrationManager.terminate(instance, timezone.now())


def log_invoice_state_transition(sender, instance, created=False, **kwargs):
    if created:
        return

    state = instance.state
    if state == models.Invoice.States.PENDING or state == instance.tracker.previous('state'):
        return

    if state == models.Invoice.States.CREATED:
        log.event_logger.invoice.info(
            'Invoice for customer {customer_name} has been created.',
            event_type='invoice_created',
            event_context={'month': instance.month, 'year': instance.year, 'customer': instance.customer}
        )
    elif state == models.Invoice.States.PAID:
        log.event_logger.invoice.info(
            'Invoice for customer {customer_name} has been paid.',
            event_type='invoice_paid',
            event_context={'month': instance.month, 'year': instance.year, 'customer': instance.customer}
        )
    elif state == models.Invoice.States.CANCELED:
        log.event_logger.invoice.info(
            'Invoice for customer {customer_name} has been canceled.',
            event_type='invoice_canceled',
            event_context={'month': instance.month, 'year': instance.year, 'customer': instance.customer}
        )


def set_tax_percent_on_invoice_creation(sender, instance, **kwargs):
    if instance.pk is not None:
        return

    payment_details = models.PaymentDetails.objects.filter(customer=instance.customer)
    if payment_details.exists():
        instance.tax_percent = payment_details.first().default_tax_percent


def set_project_name_on_invoice_item_creation(sender, instance, created=False, **kwargs):
    if created:
        item = instance
        item.project_name = item.project.name
        item.project_uuid = item.project.uuid.hex
        item.save(update_fields=('project_name', 'project_uuid'))


def update_invoice_item_on_project_name_update(sender, instance, **kwargs):
    project = instance

    if not project.tracker.has_changed('name'):
        return

    query = Q(project=project, invoice__state=models.Invoice.States.PENDING)
    for model in models.InvoiceItem.get_all_models():
        for item in model.objects.filter(query).only('pk'):
            item.project_name = project.name
            item.save(update_fields=['project_name'])


def send_invoice_report(sender, instance, created=False, **kwargs):
    if not settings.INVOICES['INVOICE_REPORTING']['ENABLE']:
        return
    invoice = instance
    if invoice.tracker.has_changed('state') and invoice.state == models.Invoice.States.CREATED:
        tasks.send_invoice_report.delay(invoice.uuid.hex)


def emit_invoice_created_event(sender, instance, created=False, **kwargs):
    if created:
        return

    state = instance.state
    if state != models.Invoice.States.CREATED or state == instance.tracker.previous('state'):
        return

    cost_signals.invoice_created.send(sender=models.Invoice,
                                      invoice=instance,
                                      issuer_details=settings.INVOICES['ISSUER_DETAILS'])




from __future__ import unicode_literals

from nodeconductor.core import NodeConductorExtension


class ExpertsExtension(NodeConductorExtension):
    class Settings:
        WALDUR_EXPERTS = {
            'REQUEST_LINK_TEMPLATE': 'https://www.example.com/#/experts/{uuid}/',
            'CONTRACT': {
                'offerings': {
                    # 'custom_vpc_experts': {
                    #     'label': 'Custom VPC',
                    #     'order': ['storage', 'ram', 'cpu_count'],
                    #     'category': 'Experts',
                    #     'description': 'Custom VPC example.',
                    #     'summary': '<div>super long long long long long long <b>summary</b></div>',
                    #     'price': 100,
                    #     'recurring_billing': False,  # False if billing is project based, True if monthly occuring.
                    #     'options': {
                    #         'storage': {
                    #             'type': 'integer',
                    #             'label': 'Max storage, GB',
                    #             'required': True,
                    #             'help_text': 'VPC storage limit in GB.',
                    #         },
                    #         'ram': {
                    #             'type': 'integer',
                    #             'label': 'Max RAM, GB',
                    #             'required': True,
                    #             'help_text': 'VPC RAM limit in GB.',
                    #         },
                    #         'cpu_count': {
                    #             'type': 'integer',
                    #             'label': 'Max vCPU',
                    #             'required': True,
                    #             'help_text': 'VPC CPU count limit.',
                    #         },
                    #     },
                    # },
                },
                # 'order': ['objectives', 'milestones', 'terms-and-conditions'],
                # 'options': {
                #     'objectives': {
                #         'order': ['objectives', 'price'],
                #         'label': 'Objectives',
                #         'description': 'Contract objectives.',
                #         'options': {
                #             'objectives': {
                #                 'type': 'text',
                #                 'label': 'Objectives',
                #                 'required': True,
                #                 'default': 'This is an objective.',
                #             },
                #             'price': {
                #                 'type': 'money',
                #                 'label': 'Planned budget',
                #             }
                #         }
                #     },
                #     'milestones': {
                #         'order': ['milestones'],
                #         'label': 'Milestones',
                #         'options': {
                #             'milestones': {
                #                 'type': 'html_text',
                #                 'label': 'Milestones',
                #                 'help_text': 'Defines project milestones.',
                #             }
                #         }
                #     },
                #     'terms-and-conditions': {
                #         'order': ['contract_methodology', 'out_of_scope', 'common_tos'],
                #         'label': 'Terms and conditions',
                #         'options': {
                #             'contract_methodology': {
                #                 'type': 'string',
                #                 'label': 'Contract methodology',
                #             },
                #             'out_of_scope': {
                #                 'type': 'string',
                #                 'label': 'Out of scope',
                #             },
                #             'common_tos': {
                #                 'type': 'text',
                #                 'label': 'Common Terms of Services.',
                #             }
                #         }
                #     },
                # }
            }
        }

    @staticmethod
    def django_app():
        return 'nodeconductor_assembly_waldur.experts'

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def celery_tasks():
        return {}



from django.db.models import Sum
from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers

from nodeconductor.core import serializers as core_serializers
from nodeconductor.core import signals as core_signals
from nodeconductor.structure import models as structure_models
from nodeconductor.structure import serializers as structure_serializers

from . import models


class PriceEstimateSerializer(serializers.HyperlinkedModelSerializer):
    scope = core_serializers.GenericRelatedField(
        related_models=models.PriceEstimate.get_estimated_models(),
        required=False
    )
    scope_name = serializers.ReadOnlyField(source='scope.name')
    scope_uuid = serializers.ReadOnlyField(source='scope.uuid')

    class Meta(object):
        model = models.PriceEstimate
        fields = ('url', 'uuid', 'scope', 'scope_name', 'scope_uuid',
                  'limit', 'total', 'threshold')
        read_only_fields = ('total', 'scope')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    def validate(self, attrs):
        if 'limit' in attrs:
            if isinstance(self.instance.scope, structure_models.Project):
                self._validate_project_limit(self.instance.scope, attrs['limit'])

            if isinstance(self.instance.scope, structure_models.Customer):
                self._validate_customer_limit(self.instance.scope, attrs['limit'])

        return attrs

    def _validate_project_limit(self, project, limit):
        customer = project.customer

        customer_limit = self._get_customer_limit(customer)
        if customer_limit == -1:
            return

        total_limit = self._get_total_limit(customer.projects.exclude(uuid=project.uuid)) + limit

        if total_limit > customer_limit:
            message = _('Total price limits of projects exceeds organization price limit. '
                        'Total limit: %(total_limit)s, organization limit: %(customer_limit)s')
            context = dict(total_limit=total_limit, customer_limit=customer_limit)
            raise serializers.ValidationError({'limit': message % context})

    def _validate_customer_limit(self, customer, limit):
        if limit == -1:
            return

        total_limit = self._get_total_limit(customer.projects.all())

        if limit < total_limit:
            message = _('Organization limit cannot be less than a sum of its projects limits: %d')
            raise serializers.ValidationError({'limit': message % total_limit})

    def _get_customer_limit(self, customer):
        try:
            estimate = models.PriceEstimate.objects.get(scope=customer)
            return estimate.limit
        except models.PriceEstimate.DoesNotExist:
            return -1

    def _get_total_limit(self, projects):
        if not projects.exists():
            return 0
        estimates = models.PriceEstimate.objects.filter(scope__in=projects).exclude(limit=-1)
        return estimates.aggregate(Sum('limit'))['limit__sum'] or 0


class NestedPriceEstimateSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.PriceEstimate
        fields = ('url', 'threshold', 'total', 'limit')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'billing-price-estimate-detail'},
        }


def get_price_estimate(serializer, scope):
    try:
        estimate = models.PriceEstimate.objects.get(scope=scope)
    except models.PriceEstimate.DoesNotExist:
        return {
            'threshold': 0.0,
            'total': 0.0,
            'limit': -1.0
        }
    else:
        serializer = NestedPriceEstimateSerializer(instance=estimate, context=serializer.context)
        return serializer.data


def add_price_estimate(sender, fields, **kwargs):
    fields['billing_price_estimate'] = serializers.SerializerMethodField()
    setattr(sender, 'get_billing_price_estimate', get_price_estimate)


core_signals.pre_serializer_fields.connect(
    sender=structure_serializers.ProjectSerializer,
    receiver=add_price_estimate,
)
core_signals.pre_serializer_fields.connect(
    sender=structure_serializers.CustomerSerializer,
    receiver=add_price_estimate,
)



from calendar import monthrange
from django.utils import timezone

from nodeconductor.core import utils as core_utils


def get_current_month():
    return timezone.now().month


def get_current_year():
    return timezone.now().year


def get_current_month_end():
    return core_utils.month_end(timezone.now())


def get_current_month_start():
    return core_utils.month_start(timezone.now())


def get_full_days(start, end):
    seconds_in_day = 24 * 60 * 60
    full_days, extra_seconds = divmod((end - start).total_seconds(), seconds_in_day)
    if extra_seconds > 0:
        full_days += 1

    return int(full_days)


def get_current_month_days():
    now = timezone.now()
    range = monthrange(now.year, now.month)
    return range[1]



from __future__ import unicode_literals

from nodeconductor_assembly_waldur.packages import views


def register_in(router):
    router.register(r'package-templates', views.PackageTemplateViewSet, base_name='package-template')
    router.register(r'openstack-packages', views.OpenStackPackageViewSet, base_name='openstack-package')



from django.db import models
from django.db.models import Q

from nodeconductor.structure import models as structure_models


class ExpertRequestQuerySet(models.QuerySet):
    def filtered_for_user(self, user):
        return self.filter(
            project__customer__permissions__is_active=True,
            project__customer__permissions__user=user,
            project__customer__permissions__role=structure_models.CustomerRole.OWNER,
        )


class ExpertRequestManager(models.Manager):
    def get_queryset(self):
        return ExpertRequestQuerySet(self.model, using=self._db)


class ExpertBidQuerySet(models.QuerySet):
    def filtered_for_user(self, user):
        expert_manager_query = Q(
            team__customer__permissions__is_active=True,
            team__customer__permissions__user=user,
            team__customer__permissions__role=structure_models.CustomerRole.OWNER,
        )
        request_customer_query = Q(
            request__project__customer__permissions__is_active=True,
            request__project__customer__permissions__user=user,
            request__project__customer__permissions__role=structure_models.CustomerRole.OWNER,
        )

        if not user.is_staff:
            return self.filter(expert_manager_query | request_customer_query)
        return self


class ExpertBidManager(models.Manager):
    def get_queryset(self):
        return ExpertBidQuerySet(self.model, using=self._db)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-09-06 13:28
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import nodeconductor.core.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='PriceEstimate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', nodeconductor.core.fields.UUIDField()),
                ('threshold', models.FloatField(default=0, validators=[django.core.validators.MinValueValidator(0)])),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('total', models.FloatField(default=0, help_text='Predicted price for scope for current month.')),
                ('limit', models.FloatField(default=-1, help_text='Price limit of a scope object in current month. -1 means no limit.')),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('packages', '0007_migrate_hourly_price_to_daily'),
    ]

    operations = [
        migrations.AlterField(
            model_name='packagecomponent',
            name='type',
            field=models.CharField(max_length=50, choices=[('ram', 'RAM'), ('cores', 'Cores'), ('storage', 'Storage')]),
        ),
    ]



from django.db import models as django_models

from nodeconductor.core import managers as core_managers
from nodeconductor.cost_tracking import managers as cost_managers


class PriceEstimateManager(core_managers.GenericKeyMixin,
                           cost_managers.UserFilterMixin,
                           django_models.Manager):

    def get_available_models(self):
        return self.model.get_estimated_models()



from . import utils


def format_event(tags):
    return {
        'measurement': 'events',
        'tags': tags,
        # fields is required, but empty
        'fields': {
            'value': 0
        }
    }


def log_resource_created(sender, instance, **kwargs):
    client = utils.get_influxdb_client()
    if not client:
        return
    title = 'Resource {resource_name} has been created.'.format(resource_name=instance.full_name)
    point = format_event({
        'title': title,
        'type': 'resource_created',
    })
    utils.write_points(client, [point])


def log_resource_deleted(sender, instance, **kwargs):
    client = utils.get_influxdb_client()
    if not client:
        return
    title = 'Resource {resource_name} has been deleted.'.format(resource_name=instance.full_name)
    point = format_event({
        'title': title,
        'type': 'resource_deleted',
    })
    utils.write_points(client, [point])



from __future__ import unicode_literals

from django.conf import settings
from django.db import transaction
from django.db.models import Q
from django.utils.translation import ugettext_lazy as _
from rest_framework import decorators, exceptions, permissions, status, response, viewsets
from django_filters.rest_framework import DjangoFilterBackend

from nodeconductor.core import views as core_views
from nodeconductor.structure import filters as structure_filters
from nodeconductor.structure import models as structure_models
from nodeconductor.structure import permissions as structure_permissions
from nodeconductor.structure import views as structure_views
from nodeconductor.users import models as user_models
from nodeconductor.users import tasks as user_tasks
from nodeconductor_assembly_waldur.support import backend as support_backend

from . import serializers, models, filters, tasks


def is_expert_manager(user):
    if user.is_staff:
        return True

    return models.ExpertProvider.objects.filter(
        customer__permissions__is_active=True,
        customer__permissions__user=user,
        customer__permissions__role=structure_models.CustomerRole.OWNER,
    ).exists()


def create_team_invitations(team, project, current_user):
    for permission in team.permissions.filter(is_active=True):
        invitation = user_models.Invitation.objects.create(
            created_by=current_user,
            email=permission.user.email,
            customer=project.customer,
            project=project,
            project_role=structure_models.ProjectRole.ADMINISTRATOR,
        )
        user_tasks.send_invitation.delay(invitation.uuid.hex, current_user.full_name or current_user.username)


def cancel_team_invitations(team, project):
    user_emails = [permission.user.email for permission in team.permissions.filter(is_active=True)]
    invitations = user_models.Invitation.objects.filter(
        email__in=user_emails,
        project=project,
        project_role=structure_models.ProjectRole.ADMINISTRATOR,
        state=user_models.Invitation.State.PENDING,
    )
    for invitation in invitations:
        invitation.cancel()


def revoke_request_permissions(expert_request):
    if not hasattr(expert_request, 'contract'):
        return
    team = expert_request.contract.team
    project = expert_request.project
    for permission in team.permissions.filter(is_active=True):
        project.remove_user(permission.user)
    cancel_team_invitations(team, project)


class ExpertProviderViewSet(viewsets.ModelViewSet):
    queryset = models.ExpertProvider.objects.all()
    serializer_class = serializers.ExpertProviderSerializer
    lookup_field = 'uuid'
    permission_classes = (permissions.IsAuthenticated,)
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.ExpertProviderFilter


class ExpertRequestViewSet(core_views.ActionsViewSet):
    queryset = models.ExpertRequest.objects.all()
    serializer_class = serializers.ExpertRequestSerializer
    lookup_field = 'uuid'
    filter_backends = (structure_filters.GenericRoleFilter, DjangoFilterBackend)
    filter_class = filters.ExpertRequestFilter
    disabled_actions = ['destroy']

    def get_queryset(self):
        qs = super(ExpertRequestViewSet, self).get_queryset()

        if not is_expert_manager(self.request.user):
            qs = qs.filtered_for_user(self.request.user)
        return qs

    @decorators.list_route()
    def configured(self, request):
        return response.Response(settings.WALDUR_EXPERTS['CONTRACT'], status=status.HTTP_200_OK)

    @transaction.atomic()
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        expert_request = serializer.save()
        support_backend.get_active_backend().create_issue(expert_request.issue)
        return response.Response(serializer.data, status=status.HTTP_201_CREATED)

    @transaction.atomic()
    @decorators.detail_route(methods=['post'])
    def cancel(self, request, *args, **kwargs):
        expert_request = self.get_object()
        expert_request.state = models.ExpertRequest.States.CANCELLED
        expert_request.save(update_fields=['state'])
        revoke_request_permissions(expert_request)
        return response.Response({'status': _('Expert request has been cancelled.')}, status=status.HTTP_200_OK)

    @transaction.atomic()
    @decorators.detail_route(methods=['post'])
    def complete(self, request, *args, **kwargs):
        expert_request = self.get_object()
        expert_request.state = models.ExpertRequest.States.COMPLETED
        expert_request.save(update_fields=['state'])
        revoke_request_permissions(expert_request)
        return response.Response({'status': _('Expert request has been completed.')}, status=status.HTTP_200_OK)

    def is_valid_request(request, view, obj=None):
        expert_request = obj

        if not expert_request:
            return

        if expert_request.state not in (models.ExpertRequest.States.ACTIVE, models.ExpertRequest.States.PENDING):
            raise exceptions.ValidationError(_('Expert request should be in active or pending state.'))

    def is_owner(request, view, obj=None):
        expert_request = obj

        if not expert_request:
            return

        if not structure_permissions._has_owner_access(request.user, expert_request.project.customer):
            raise exceptions.PermissionDenied()

    cancel_permissions = complete_permissions = [is_owner, is_valid_request]


class ExpertBidViewSet(core_views.ActionsViewSet):
    queryset = models.ExpertBid.objects.all()
    serializer_class = serializers.ExpertBidSerializer
    lookup_field = 'uuid'
    filter_backends = (DjangoFilterBackend,)
    filter_class = filters.ExpertBidFilter
    disabled_actions = ['update']

    def is_expert_manager(request, view, obj=None):
        if not is_expert_manager(request.user):
            raise exceptions.PermissionDenied()

    create_permissions = [is_expert_manager]

    def get_queryset(self):
        return super(ExpertBidViewSet, self).get_queryset()\
            .filtered_for_user(self.request.user).distinct()

    @decorators.detail_route(methods=['post'])
    @transaction.atomic()
    def accept(self, request, uuid=None):
        current_user = self.request.user
        expert_bid = self.get_object()

        expert_request = expert_bid.request
        expert_request.state = models.ExpertRequest.States.ACTIVE
        expert_request.save(update_fields=['state'])

        models.ExpertContract.objects.create(
            request=expert_request,
            team=expert_bid.team,
            price=expert_bid.price,
            description=expert_bid.description,
        )

        create_team_invitations(expert_bid.team, expert_request.project, current_user)

        for permission in expert_request.project.customer.permissions.all():
            tasks.send_contract.delay(expert_request.uuid.hex, permission.user.email)

        return response.Response({'status': _('Expert bid has been accepted.')}, status=status.HTTP_200_OK)

    def is_pending_request(request, view, obj=None):
        if obj and obj.request.state != models.ExpertRequest.States.PENDING:
            raise exceptions.ValidationError(_('Expert request should be in pending state.'))

    accept_permissions = [structure_permissions.is_owner, is_pending_request]

    def can_delete_bid(request, view, obj=None):
        if not obj:
            return
        if request.user.is_staff:
            return
        if not structure_permissions._has_owner_access(request.user, obj.team.customer):
            raise exceptions.PermissionDenied()

    destroy_permissions = [is_pending_request, can_delete_bid]


def get_project_experts_count(project):
    valid_states = (models.ExpertRequest.States.ACTIVE, models.ExpertRequest.States.PENDING)
    query = Q(project=project, state__in=valid_states)
    return models.ExpertRequest.objects.filter(query).count()


def get_customer_experts_count(customer):
    query = Q(state=models.ExpertRequest.States.PENDING) |\
            Q(state=models.ExpertRequest.States.ACTIVE, contract__team__customer=customer)
    return models.ExpertRequest.objects.filter(query).count()


structure_views.ProjectCountersView.register_counter('experts', get_project_experts_count)
structure_views.CustomerCountersView.register_counter('experts', get_customer_experts_count)



import datetime
from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from nodeconductor.core import serializers as core_serializers
from nodeconductor.core import signals as core_signals
from nodeconductor.core import utils as core_utils
from nodeconductor.structure import SupportedServices
from nodeconductor.structure import serializers as structure_serializers
from nodeconductor_assembly_waldur.common.utils import quantize_price
from nodeconductor_openstack.openstack import models as openstack_models
from nodeconductor_assembly_waldur.support import models as support_models

from . import models


class InvoiceItemSerializer(serializers.HyperlinkedModelSerializer):
    tax = serializers.DecimalField(max_digits=15, decimal_places=7)
    total = serializers.DecimalField(max_digits=15, decimal_places=7)

    scope_type = serializers.SerializerMethodField()
    scope_uuid = serializers.SerializerMethodField()

    class Meta(object):
        model = models.InvoiceItem
        fields = ('name', 'price', 'tax', 'total', 'unit_price', 'unit',
                  'start', 'end', 'product_code', 'article_code', 'project_name', 'project_uuid',
                  'scope_type', 'scope_uuid',)

    def get_scope_type(self, item):
        # It should be implemented by inherited class
        return

    def get_scope_uuid(self, item):
        # It should be implemented by inherited class
        return


class OpenStackItemSerializer(InvoiceItemSerializer):
    tenant_name = serializers.ReadOnlyField(source='get_tenant_name')
    tenant_uuid = serializers.ReadOnlyField(source='get_tenant_uuid')
    template_name = serializers.ReadOnlyField(source='get_template_name')
    template_uuid = serializers.ReadOnlyField(source='get_template_uuid')
    template_category = serializers.ReadOnlyField(source='get_template_category')

    class Meta(InvoiceItemSerializer.Meta):
        model = models.OpenStackItem
        fields = InvoiceItemSerializer.Meta.fields + ('package', 'tenant_name', 'tenant_uuid', 'usage_days',
                                                      'template_name', 'template_uuid', 'template_category')
        extra_kwargs = {
            'package': {'lookup_field': 'uuid', 'view_name': 'openstack-package-detail'},
        }

    def get_scope_type(self, item):
        return SupportedServices.get_name_for_model(openstack_models.Tenant)

    def get_scope_uuid(self, item):
        return item.get_tenant_uuid()


class OfferingItemSerializer(InvoiceItemSerializer):
    offering_type = serializers.ReadOnlyField(source='get_offering_type')

    class Meta(InvoiceItemSerializer.Meta):
        model = models.OfferingItem
        fields = InvoiceItemSerializer.Meta.fields + ('offering', 'offering_type', 'usage_days')
        extra_kwargs = {
            'offering': {'lookup_field': 'uuid', 'view_name': 'support-offering-detail'},
        }

    def get_scope_type(self, item):
        return support_models.Offering.get_scope_type()

    def get_scope_uuid(self, item):
        return item.get_offering_uuid()


class GenericItemSerializer(InvoiceItemSerializer):

    class Meta(InvoiceItemSerializer.Meta):
        model = models.GenericInvoiceItem
        fields = InvoiceItemSerializer.Meta.fields + ('quantity',)

    def get_scope_type(self, item):
        return SupportedServices.get_name_for_model(item.content_type.model_class())

    def get_scope_uuid(self, item):
        if item.scope:
            return item.scope.uuid.hex
        return item.details['scope_uuid']


class InvoiceSerializer(core_serializers.RestrictedSerializerMixin,
                        serializers.HyperlinkedModelSerializer):
    price = serializers.DecimalField(max_digits=15, decimal_places=7)
    tax = serializers.DecimalField(max_digits=15, decimal_places=7)
    total = serializers.DecimalField(max_digits=15, decimal_places=7)
    openstack_items = OpenStackItemSerializer(many=True)
    offering_items = OfferingItemSerializer(many=True)
    generic_items = GenericItemSerializer(many=True)
    issuer_details = serializers.SerializerMethodField()
    customer_details = serializers.SerializerMethodField()
    due_date = serializers.DateField()

    class Meta(object):
        model = models.Invoice
        fields = (
            'url', 'uuid', 'number', 'customer', 'price', 'tax', 'total',
            'state', 'year', 'month', 'issuer_details', 'invoice_date', 'due_date',
            'customer', 'customer_details',
            'openstack_items', 'offering_items', 'generic_items',
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'customer': {'lookup_field': 'uuid'},
        }

    def get_issuer_details(self, invoice):
        return settings.INVOICES['ISSUER_DETAILS']

    def get_customer_details(self, invoice):
        try:
            payment_details = models.PaymentDetails.objects.get(customer=invoice.customer)
        except models.PaymentDetails.DoesNotExist:
            return
        return {
            'company': payment_details.company,
            'address': payment_details.address,
            'country': payment_details.country,
            'email': payment_details.email,
            'postal': payment_details.postal,
            'phone': payment_details.phone,
            'bank': payment_details.bank,
        }


class InvoiceNotificationSerializer(serializers.Serializer):
    link_template = serializers.URLField(help_text=_('The template must include {uuid} parameter '
                                                     'e.g. http://example.com/invoice/{uuid}'))

    def validate_link_template(self, link_template):
        if '{uuid}' not in link_template:
            raise serializers.ValidationError(_("Link template must include '{uuid}' parameter."))

        return link_template


class PaymentDetailsSerializer(core_serializers.AugmentedSerializerMixin,
                               serializers.HyperlinkedModelSerializer):

    type = serializers.ChoiceField(choices=[(t, t) for t in settings.INVOICES['COMPANY_TYPES']],
                                   allow_blank=True,
                                   required=False)

    class Meta(object):
        model = models.PaymentDetails
        fields = (
            'url', 'uuid', 'customer', 'company', 'type', 'address',
            'country', 'email', 'postal', 'phone', 'bank', 'account',
            'default_tax_percent', 'accounting_start_date', 'is_billable',
        )
        protected_fields = ('customer',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'customer': {'lookup_field': 'uuid'},
        }

    def get_fields(self):
        fields = super(PaymentDetailsSerializer, self).get_fields()
        if isinstance(self.instance, models.PaymentDetails):
            fields['accounting_start_date'].read_only = self.instance.is_billable()
        return fields


class InvoiceItemReportSerializer(serializers.ModelSerializer):
    invoice_number = serializers.ReadOnlyField(source='invoice.number')
    invoice_uuid = serializers.ReadOnlyField(source='invoice.uuid')
    invoice_year = serializers.ReadOnlyField(source='invoice.year')
    invoice_month = serializers.ReadOnlyField(source='invoice.month')
    invoice_date = serializers.ReadOnlyField(source='invoice.invoice_date')
    due_date = serializers.ReadOnlyField(source='invoice.due_date')
    customer_uuid = serializers.ReadOnlyField(source='invoice.customer.uuid')
    customer_name = serializers.ReadOnlyField(source='invoice.customer.name')

    class Meta(object):
        model = models.OpenStackItem
        fields = (
            'customer_uuid', 'customer_name',
            'project_uuid', 'project_name',
            'invoice_uuid', 'invoice_number',
            'invoice_year', 'invoice_month',
            'invoice_date', 'due_date',
            'invoice_price', 'invoice_tax', 'invoice_total',
            'name', 'article_code', 'product_code',
            'price', 'tax', 'total', 'unit_price', 'unit',
            'start', 'end',
        )
        decimal_fields = (
            'price', 'tax', 'total', 'unit_price',
            'invoice_price', 'invoice_tax', 'invoice_total'
        )
        decimal_fields_extra_kwargs = {
            'invoice_price': {
                'source': 'invoice.price',
            },
            'invoice_tax': {
                'source': 'invoice.tax',
            },
            'invoice_total': {
                'source': 'invoice.total',
            },
        }

    def build_field(self, field_name, info, model_class, nested_depth):
        if field_name in self.Meta.decimal_fields:
            field_class = serializers.DecimalField
            field_kwargs = dict(
                max_digits=20,
                decimal_places=2,
                coerce_to_string=True,
            )
            default_kwargs = self.Meta.decimal_fields_extra_kwargs.get(field_name)
            if default_kwargs:
                field_kwargs.update(default_kwargs)
            return field_class, field_kwargs

        return super(InvoiceItemReportSerializer, self).build_field(field_name, info, model_class, nested_depth)

    def get_extra_kwargs(self):
        extra_kwargs = super(InvoiceItemReportSerializer, self).get_extra_kwargs()
        extra_kwargs.update(settings.INVOICES['INVOICE_REPORTING']['SERIALIZER_EXTRA_KWARGS'])
        return extra_kwargs


class OpenStackItemReportSerializer(InvoiceItemReportSerializer):
    class Meta(InvoiceItemReportSerializer.Meta):
        model = models.OpenStackItem
        fields = InvoiceItemReportSerializer.Meta.fields + ('usage_days',)


class OfferingItemReportSerializer(InvoiceItemReportSerializer):
    class Meta(InvoiceItemReportSerializer.Meta):
        model = models.OfferingItem
        fields = InvoiceItemReportSerializer.Meta.fields + ('usage_days',)


class GenericItemReportSerializer(InvoiceItemReportSerializer):
    class Meta(InvoiceItemReportSerializer.Meta):
        model = models.GenericInvoiceItem
        fields = InvoiceItemReportSerializer.Meta.fields + ('quantity',)


# SAF is accounting soft from Estonia: www.sysdec.ee/safsaf.htm
class SAFReportSerializer(serializers.Serializer):
    KUUPAEV = serializers.SerializerMethodField(method_name='get_last_day_of_month')
    VORMKUUP = serializers.SerializerMethodField(method_name='get_invoice_date')
    MAKSEAEG = serializers.SerializerMethodField(method_name='get_due_date')
    YKSUS = serializers.ReadOnlyField(source='invoice.customer.agreement_number')
    PARTNER = serializers.ReadOnlyField(source='invoice.customer.agreement_number')
    ARTIKKEL = serializers.ReadOnlyField(source='article_code')
    KOGUS = serializers.SerializerMethodField(method_name='get_quantity')
    SUMMA = serializers.SerializerMethodField(method_name='get_total')
    RMAKSUSUM = serializers.SerializerMethodField(method_name='get_tax')
    RMAKSULIPP = serializers.SerializerMethodField(method_name='get_vat')
    ARTPROJEKT = serializers.SerializerMethodField(method_name='get_project')
    ARTNIMI = serializers.ReadOnlyField(source='name')
    VALI = serializers.SerializerMethodField(method_name='get_empty_field')
    U_KONEDEARV = serializers.SerializerMethodField(method_name='get_empty_field')
    H_PERIOOD = serializers.SerializerMethodField(method_name='get_covered_period')

    class Meta(object):
        fields = ('KUUPAEV', 'VORMKUUP', 'MAKSEAEG', 'YKSUS', 'PARTNER',
                  'ARTIKKEL', 'KOGUS', 'SUMMA', 'RMAKSUSUM', 'RMAKSULIPP',
                  'ARTPROJEKT', 'ARTNIMI', 'VALI', 'U_KONEDEARV', 'H_PERIOOD')

    def format_date(self, date):
        if date:
            return date.strftime('%d.%m.%Y')
        return ''

    def get_first_day(self, invoice_item):
        year = invoice_item.invoice.year
        month = invoice_item.invoice.month
        return datetime.date(year=year, month=month, day=1)

    def get_last_day_of_month(self, invoice_item):
        first_day = self.get_first_day(invoice_item)
        last_day = core_utils.month_end(first_day)
        return self.format_date(last_day)

    def get_invoice_date(self, invoice_item):
        date = invoice_item.invoice.invoice_date
        return self.format_date(date)

    def get_due_date(self, invoice_item):
        date = invoice_item.invoice.due_date
        return self.format_date(date)

    def get_quantity(self, invoice_item):
        if hasattr(invoice_item, 'quantity'):
            return invoice_item.quantity
        return invoice_item.usage_days

    def get_total(self, invoice_item):
        return quantize_price(invoice_item.total)

    def get_tax(self, invoice_item):
        return quantize_price(invoice_item.tax)

    def get_project(self, invoice_item):
        return settings.INVOICES['INVOICE_REPORTING']['SAF_PARAMS']['ARTPROJEKT']

    def get_vat(self, invoice_item):
        return settings.INVOICES['INVOICE_REPORTING']['SAF_PARAMS']['RMAKSULIPP']

    def get_empty_field(self, invoice_item):
        return ''

    def get_covered_period(self, invoice_item):
        first_day = self.get_first_day(invoice_item)
        last_day = core_utils.month_end(first_day)
        return '%s-%s' % (self.format_date(first_day), self.format_date(last_day))


def get_accounting_start_date(serializer, scope):
    if not hasattr(scope, 'payment_details'):
        return None
    return scope.payment_details.accounting_start_date


def add_accounting_start_date(sender, fields, **kwargs):
    fields['accounting_start_date'] = serializers.SerializerMethodField()
    setattr(sender, 'get_accounting_start_date', get_accounting_start_date)


core_signals.pre_serializer_fields.connect(
    sender=structure_serializers.CustomerSerializer,
    receiver=add_accounting_start_date,
)



import copy
import itertools

from django.conf import settings
from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions, serializers

from nodeconductor.core import models as core_models
from nodeconductor.core import serializers as core_serializers
from nodeconductor.core import signals as core_signals
from nodeconductor.structure import permissions as structure_permissions
from nodeconductor.structure import serializers as structure_serializers
from nodeconductor_assembly_waldur.support import models as support_models
from nodeconductor_assembly_waldur.support import serializers as support_serializers

from . import models


class ExpertProviderSerializer(core_serializers.AugmentedSerializerMixin,
                               serializers.HyperlinkedModelSerializer):
    agree_with_policy = serializers.BooleanField(write_only=True, required=False)

    class Meta(object):
        model = models.ExpertProvider
        fields = ('url', 'uuid', 'created', 'customer', 'customer_name', 'agree_with_policy')
        read_only_fields = ('url', 'uuid', 'created')
        related_paths = {
            'customer': ('uuid', 'name', 'native_name', 'abbreviation')
        }
        protected_fields = ('customer',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'expert-provider-detail'},
            'customer': {'lookup_field': 'uuid'},
        }

    def validate(self, attrs):
        agree_with_policy = attrs.pop('agree_with_policy', False)
        if not agree_with_policy:
            raise serializers.ValidationError(
                {'agree_with_policy': _('User must agree with policies to register organization.')})

        structure_permissions.is_owner(self.context['request'], None, attrs['customer'])
        return attrs


class ExpertContractSerializer(core_serializers.AugmentedSerializerMixin,
                               serializers.HyperlinkedModelSerializer):
    class Meta(object):
        model = models.ExpertContract
        fields = (
            'price', 'description', 'team', 'team_uuid', 'team_name',
        )
        related_paths = {
            'team': ('uuid', 'name'),
        }
        extra_kwargs = {
            'team': {'lookup_field': 'uuid', 'view_name': 'project-detail'},
        }


class ExpertRequestSerializer(support_serializers.ConfigurableSerializerMixin,
                              core_serializers.AugmentedSerializerMixin,
                              serializers.HyperlinkedModelSerializer):
    type = serializers.ChoiceField(choices=settings.WALDUR_EXPERTS['CONTRACT']['offerings'].keys())
    state = serializers.ReadOnlyField(source='get_state_display')
    description = serializers.CharField(required=False)
    contract = ExpertContractSerializer(required=False, read_only=True)
    customer = serializers.HyperlinkedRelatedField(
        source='project.customer',
        view_name='customer-detail',
        read_only=True,
        lookup_field='uuid'
    )
    customer_name = serializers.ReadOnlyField(source='project.customer.name')
    customer_uuid = serializers.ReadOnlyField(source='project.customer.uuid')
    type_label = serializers.SerializerMethodField()
    extra = serializers.JSONField(read_only=True)

    def get_type_label(self, instance):
        return self._get_configuration(instance.type).get('label', instance.type)

    def _get_offerings_configuration(self):
        return copy.deepcopy(settings.WALDUR_EXPERTS['CONTRACT']['offerings'])

    def _get_configuration(self, type):
        contract_config = copy.deepcopy(settings.WALDUR_EXPERTS['CONTRACT'])
        offering_config = contract_config['offerings'].get(type)
        configured_options = [tab.get('options', {}) for tab in contract_config.get('options').values()]
        options_order = [tab.get('order', []) for tab in contract_config.get('options').values()]
        offering_config['order'] = offering_config['order'] + list(itertools.chain(*options_order))
        [offering_config['options'].update(options) for options in configured_options]

        return offering_config

    class Meta(object):
        model = models.ExpertRequest
        fields = ('url', 'uuid', 'name', 'type', 'state', 'type_label', 'description', 'extra',
                  'customer', 'customer_name', 'customer_uuid',
                  'project', 'project_name', 'project_uuid',
                  'created', 'modified', 'contract', 'recurring_billing',
                  'objectives', 'milestones', 'contract_methodology', 'out_of_scope', 'common_tos',
                  'issue', 'issue_name', 'issue_link', 'issue_key', 'issue_description', 'issue_uuid', 'issue_status',)
        read_only_fields = ('price', 'state', 'issue', 'recurring_billing', 'extra')
        protected_fields = ('project', 'type', 'common_tos')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'expert-request-detail'},
            'project': {'lookup_field': 'uuid', 'view_name': 'project-detail'},
            'issue': {'lookup_field': 'uuid', 'view_name': 'support-issue-detail'},
        }
        related_paths = {
            'project': ('uuid', 'name'),
            'issue': ('uuid', 'name', 'status', 'key', 'description', 'link'),
        }

    def validate_project(self, project):
        request = self.context['request']
        structure_permissions.is_owner(request, None, project.customer)
        if models.ExpertRequest.objects.filter(
            state=models.ExpertRequest.States.ACTIVE,
            project=project
        ).exists():
            raise serializers.ValidationError(_('Active expert request for current project already exists.'))
        return project

    def create(self, validated_data):
        request = self.context['request']
        project = validated_data['project']
        type = validated_data['type']

        configuration = self._get_configuration(type)
        type_label = configuration.get('label', type)
        issue_details = dict(
            caller=request.user,
            project=project,
            customer=project.customer,
            type=settings.WALDUR_SUPPORT['DEFAULT_OFFERING_ISSUE_TYPE'],
            description=self._form_description(configuration, validated_data),
            summary='Request for \'%s\'' % type_label)
        issue_details['summary'] = support_serializers.render_issue_template('summary', issue_details)
        issue_details['description'] = support_serializers.render_issue_template('description', issue_details)
        issue = support_models.Issue.objects.create(**issue_details)

        return models.ExpertRequest.objects.create(
            user=request.user,
            project=project,
            name=validated_data.get('name'),
            type=type,
            recurring_billing=configuration.get('recurring_billing', False),
            description=issue_details['description'],
            extra=self._get_extra(configuration, validated_data),
            issue=issue,
            objectives=validated_data.get('objectives', ''),
            milestones=validated_data.get('milestones', ''),
            contract_methodology=validated_data.get('contract_methodology', ''),
            out_of_scope=validated_data.get('out_of_scope', ''),
            common_tos=validated_data.get('common_tos', ''),
        )


class ExpertBidSerializer(core_serializers.AugmentedSerializerMixin,
                          serializers.HyperlinkedModelSerializer):

    team_members = serializers.SerializerMethodField(source='get_team_members')
    customer_uuid = serializers.ReadOnlyField(source='team.customer.uuid')
    customer_name = serializers.ReadOnlyField(source='team.customer.name')
    customer_email = serializers.ReadOnlyField(source='team.customer.email')

    class Meta(object):
        model = models.ExpertBid
        fields = (
            'url', 'uuid', 'created', 'modified', 'price', 'description',
            'team', 'team_uuid', 'team_name', 'team_members',
            'request', 'request_uuid', 'request_name',
            'customer_uuid', 'customer_name', 'customer_email',
        )
        related_paths = {
            'team': ('uuid', 'name'),
            'request': ('uuid', 'name'),
        }
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'expert-bid-detail'},
            'request': {'lookup_field': 'uuid', 'view_name': 'expert-request-detail'},
            'team': {'lookup_field': 'uuid', 'view_name': 'project-detail'},
        }

    def validate_request(self, request):
        if request.state != models.ExpertRequest.States.PENDING:
            raise exceptions.ValidationError(_('Expert request should be in pending state.'))
        return request

    def validate_team(self, team):
        if not team.permissions.filter(is_active=True).exists():
            raise exceptions.ValidationError(_('Expert team should have at least one member.'))

        return team

    def validate(self, attrs):
        team = attrs['team']
        request = attrs['request']

        if models.ExpertBid.objects.filter(request=request, team__customer=team.customer).exists():
            raise exceptions.ValidationError({'team': _('There is a bid from this customer already.')})

        return attrs

    def create(self, validated_data):
        request = self.context['request']
        validated_data['user'] = request.user
        return super(ExpertBidSerializer, self).create(validated_data)

    def get_team_members(self, request):
        user_ids = request.team.permissions.filter(is_active=True).values_list('user_id')
        users = core_models.User.objects.filter(pk__in=user_ids)
        return users.values('username', 'email', 'full_name', 'uuid')


def get_is_expert_provider(serializer, scope):
    customer = structure_permissions._get_customer(scope)
    return models.ExpertProvider.objects.filter(customer=customer).exists()


def add_expert_provider(sender, fields, **kwargs):
    fields['is_expert_provider'] = serializers.SerializerMethodField()
    setattr(sender, 'get_is_expert_provider', get_is_expert_provider)


core_signals.pre_serializer_fields.connect(
    sender=structure_serializers.CustomerSerializer,
    receiver=add_expert_provider,
)
core_signals.pre_serializer_fields.connect(
    sender=structure_serializers.CustomerPermissionSerializer,
    receiver=add_expert_provider,
)



from decimal import Decimal, ROUND_UP


def quantize_price(value):
    """
    Returns value rounded up to 2 places after the decimal point.
    :rtype: Decimal
    """
    return value.quantize(Decimal('0.01'), rounding=ROUND_UP)



from django.apps import AppConfig
from django.db.models import signals


class AnalyticsConfig(AppConfig):
    name = 'nodeconductor_assembly_waldur.analytics'
    verbose_name = 'Analytics'

    def ready(self):
        from nodeconductor.structure.models import ResourceMixin
        from . import handlers

        for index, model in enumerate(ResourceMixin.get_all_models()):
            signals.post_save.connect(
                handlers.log_resource_created,
                sender=model,
                dispatch_uid=('nodeconductor_assembly_waldur.analytics.handlers.'
                              'log_resource_created_{}_{}'.format(model.__name__, index)),
            )

            signals.post_delete.connect(
                handlers.log_resource_deleted,
                sender=model,
                dispatch_uid=('nodeconductor_assembly_waldur.analytics.handlers.'
                              'log_resource_deleted_{}_{}'.format(model.__name__, index)),
            )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0002_comment_and_support_user'),
    ]

    operations = [
        migrations.AddField(
            model_name='issue',
            name='impact',
            field=models.CharField(max_length=255, blank=True),
        ),
        migrations.AddField(
            model_name='issue',
            name='link',
            field=models.URLField(help_text='Link to issue in support system.', max_length=255, blank=True),
        ),
        migrations.AddField(
            model_name='issue',
            name='priority',
            field=models.CharField(max_length=255, blank=True),
        ),
        migrations.AlterField(
            model_name='issue',
            name='type',
            field=models.CharField(max_length=255),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.1 on 2017-09-22 08:49
from __future__ import unicode_literals

from django.db import migrations, models
import nodeconductor.core.fields
import nodeconductor.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('slurm_invoices', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='slurmpackage',
            name='name',
            field=models.CharField(default='default', max_length=150, validators=[nodeconductor.core.validators.validate_name], verbose_name='name'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='slurmpackage',
            name='uuid',
            field=nodeconductor.core.fields.UUIDField(),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from decimal import Decimal
import django.core.validators

from .. import utils


class Migration(migrations.Migration):

    dependencies = [
        ('invoices', '0005_add_company_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='openstackitem',
            name='daily_price',
            field=models.DecimalField(default=0, help_text='Price per day.', max_digits=22, decimal_places=7, validators=[django.core.validators.MinValueValidator(Decimal('0'))]),
        ),
    ]



from __future__ import unicode_literals

from decimal import Decimal
from django.conf import settings
from django.core.validators import MinValueValidator
from django.utils.translation import ugettext_lazy as _
from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from model_utils import FieldTracker

from nodeconductor.core import fields as core_fields
from nodeconductor.core import models as core_models
from nodeconductor.structure import models as structure_models
from nodeconductor_assembly_waldur.support import models as support_models
from nodeconductor_assembly_waldur.common import mixins as common_mixins

from . import managers


@python_2_unicode_compatible
class ExpertProvider(core_models.UuidMixin,
                     structure_models.TimeStampedModel):
    customer = models.OneToOneField(structure_models.Customer, related_name='+', on_delete=models.CASCADE)

    class Meta(object):
        verbose_name = _('Expert providers')

    def __str__(self):
        return str(self.customer)

    @classmethod
    def get_url_name(cls):
        return 'expert-provider'


class PriceMixin(models.Model):
    class Meta(object):
        abstract = True

    price = models.DecimalField(max_digits=22, decimal_places=7,
                                validators=[MinValueValidator(Decimal('0'))],
                                default=0)


@python_2_unicode_compatible
class ExpertRequest(core_models.UuidMixin,
                    core_models.NameMixin,
                    core_models.DescribableMixin,
                    PriceMixin,
                    common_mixins.ProductCodeMixin,
                    structure_models.StructureLoggableMixin,
                    structure_models.TimeStampedModel):
    class States(object):
        PENDING = 'pending'
        ACTIVE = 'active'
        CANCELLED = 'cancelled'
        COMPLETED = 'completed'

        CHOICES = (
            (PENDING, _('Pending')),
            (ACTIVE, _('Active')),
            (CANCELLED, _('Cancelled')),
            (COMPLETED, _('Completed'))
        )

    user = models.ForeignKey(core_models.User, related_name='+', on_delete=models.CASCADE,
                             help_text=_('The user which has created this request.'))
    project = models.ForeignKey(structure_models.Project, related_name='+', on_delete=models.CASCADE)
    state = models.CharField(default=States.PENDING, max_length=30, choices=States.CHOICES)
    type = models.CharField(max_length=255)
    extra = core_fields.JSONField(default={})
    issue = models.ForeignKey(support_models.Issue, null=True, on_delete=models.SET_NULL)
    recurring_billing = models.BooleanField(
        default=False, help_text=_('Defines whether expert request has to be billed every month or only once'))
    objectives = models.TextField(blank=True)
    milestones = models.TextField(blank=True)
    contract_methodology = models.TextField(blank=True)
    out_of_scope = models.TextField(
        blank=True, help_text=_('Elements that are explicitly excluded from the contract'))
    common_tos = models.TextField(blank=True)

    tracker = FieldTracker()
    objects = managers.ExpertRequestManager()

    class Meta:
        ordering = ['-created']

    def get_log_fields(self):
        return super(ExpertRequest, self).get_log_fields() + ('state', 'project', 'user')

    @classmethod
    def get_url_name(cls):
        return 'expert-request'

    @property
    def type_label(self):
        offerings = settings.WALDUR_SUPPORT.get('CONTRACT', {}).get('offerings', {})
        type_settings = offerings.get(self.type, {})
        return type_settings.get('label', None)

    @property
    def link(self):
        return settings.WALDUR_EXPERTS['REQUEST_LINK_TEMPLATE'].format(uuid=self.uuid.hex)

    def __str__(self):
        return '{} / {}'.format(self.project.name, self.project.customer.name)


class ExpertBid(core_models.UuidMixin,
                core_models.DescribableMixin,
                PriceMixin,
                structure_models.StructureLoggableMixin,
                structure_models.TimeStampedModel):
    user = models.ForeignKey(core_models.User, related_name='+', on_delete=models.CASCADE,
                             help_text=_('The user which has created this bid.'))
    request = models.ForeignKey(ExpertRequest, on_delete=models.CASCADE, related_name='bids')
    team = models.ForeignKey(structure_models.Project, related_name='+')
    objects = managers.ExpertBidManager()
    tracker = FieldTracker()

    class Meta:
        ordering = ['-created']

    class Permissions(object):
        customer_path = 'request__project__customer'

    def get_log_fields(self):
        return super(ExpertBid, self).get_log_fields() + ('request', 'user', 'price')

    @classmethod
    def get_url_name(cls):
        return 'expert-bid'


class ExpertContract(PriceMixin, core_models.DescribableMixin, structure_models.TimeStampedModel):
    request = models.OneToOneField(ExpertRequest, on_delete=models.CASCADE, related_name='contract')
    team = models.ForeignKey(structure_models.Project, related_name='+', on_delete=models.PROTECT)

    class Meta:
        ordering = ['-created']



from __future__ import unicode_literals

from nodeconductor_assembly_waldur.slurm_invoices import views


def register_in(router):
    router.register(r'slurm-packages', views.SlurmPackageViewSet, base_name='slurm-package')



from __future__ import unicode_literals

from nodeconductor_assembly_waldur.invoices import views


def register_in(router):
    router.register(r'invoices', views.InvoiceViewSet, base_name='invoice')
    router.register(r'payment-details', views.PaymentDetailsViewSet, base_name='payment-details')



import collections

from decimal import Decimal
from django import forms
from django.conf import settings
from django.contrib import admin
from django.contrib.admin import widgets
from django.forms.models import BaseInlineFormSet
from django.utils.translation import ugettext_lazy as _

from nodeconductor.core import admin as core_admin
from nodeconductor.structure import models as structure_models
from nodeconductor_assembly_waldur.packages import models, utils


class GBtoMBWidget(widgets.AdminIntegerFieldWidget):
    def value_from_datadict(self, data, files, name):
        value = super(GBtoMBWidget, self).value_from_datadict(data, files, name) or 0
        value = int(value) * 1024
        return value

    def format_value(self, value):
        return int(value) / 1024

    def render(self, name, value, attrs=None, renderer=None):
        result = super(GBtoMBWidget, self).render(name, value, attrs)
        return '<label>%s GB</label>' % result


class PriceForMBinGBWidget(forms.NumberInput):
    def __init__(self, attrs):
        self.readonly = attrs.pop('readonly', False)
        super(PriceForMBinGBWidget, self).__init__(attrs)

    def value_from_datadict(self, data, files, name):
        value = super(PriceForMBinGBWidget, self).value_from_datadict(data, files, name) or 0
        value = Decimal(value) / 1024
        return value

    def format_value(self, value):
        return Decimal(value) * 1024

    def render(self, name, value, attrs=None, renderer=None):
        if self.readonly:
            return core_admin.render_to_readonly(self._format_value(value))
        else:
            return super(PriceForMBinGBWidget, self).render(name, value, attrs)


class PackageComponentForm(forms.ModelForm):
    monthly_price = forms.DecimalField(label=_('Price for 30 days'), initial=0, required=True)
    price = forms.DecimalField(initial=0, label=_('Price per unit per day'), required=False,
                               widget=core_admin.ReadonlyTextWidget())
    if settings.DEBUG:
        price_per_day = forms.DecimalField(label=_('Price per day for MB'),
                                           initial=0,
                                           required=False,
                                           widget=core_admin.ReadonlyTextWidget())

    class Meta:
        model = models.PackageComponent
        price_defining_fields = ('amount', 'monthly_price')
        fields = ('type', 'price') + price_defining_fields

    def __init__(self, *args, **kwargs):
        super(PackageComponentForm, self).__init__(*args, **kwargs)
        if self.instance:
            self.fields['monthly_price'].initial = self.instance.monthly_price
            if self.instance.type in models.PackageTemplate.get_memory_types():
                self.fields['price'].widget = PriceForMBinGBWidget(attrs={'readonly': True})

            if settings.DEBUG:
                self.fields['price_per_day'].initial = self.instance.price

            if hasattr(self.instance, 'template') and self.instance.template.openstack_packages.exists():
                for field in (field for field in self.fields if field in self.Meta.price_defining_fields):
                    self.fields[field].widget.attrs['readonly'] = True

    def clean(self):
        super(PackageComponentForm, self).clean()
        if 'monthly_price' not in self.cleaned_data or 'amount' not in self.cleaned_data:
            return

        instance = getattr(self, 'instance', None)
        template = getattr(instance, 'template', None)
        if template and template.openstack_packages.exists() and (
                        'monthly_price' in self.changed_data or 'amount' in self.changed_data):
            raise forms.ValidationError(_('Price cannot be changed for a template which has connected packages.'))

        type = self.cleaned_data['type']
        monthly_price = self.cleaned_data['monthly_price']
        amount = self.cleaned_data['amount']

        price_min = 10 ** -models.PackageComponent.PRICE_DECIMAL_PLACES
        monthly_price_min = price_min * 30 * amount
        if monthly_price < monthly_price_min and monthly_price != 0:
            raise forms.ValidationError(_('Monthly price for "%(type)s" should be greater than %(min)s or equal to 0') % {
                'type': type,
                'min': monthly_price_min,
            })

        price_max = 10 ** (models.PackageComponent.PRICE_MAX_DIGITS - models.PackageComponent.PRICE_DECIMAL_PLACES)
        monthly_price_max = price_max * 30 * amount
        if monthly_price > monthly_price_max:
            raise forms.ValidationError(_('Monthly price for "%(type)s" should be lower than %(max)s') % {
                'type': type,
                'max': monthly_price_max
            })

    def save(self, commit=True):
        monthly_price = self.cleaned_data.get('monthly_price', None)
        amount = self.cleaned_data.get('amount', 0)
        package_component = super(PackageComponentForm, self).save(commit=commit)
        if amount:
            package_component.price = monthly_price / 30 / amount
        if commit:
            package_component.save()
        return package_component


class PackageComponentInlineFormset(BaseInlineFormSet):
    """
    Formset responsible for package component inlines validation and their initial population.
    """
    form = PackageComponentForm

    def __init__(self, **kwargs):
        # Fill inlines with required component types
        kwargs['initial'] = [{'type': t} for t in models.PackageTemplate.get_required_component_types()]
        super(PackageComponentInlineFormset, self).__init__(**kwargs)

    def add_fields(self, form, index):
        super(PackageComponentInlineFormset, self).add_fields(form, index)
        if 'type' in form.initial and form.initial['type'] in models.PackageTemplate.get_memory_types():
            is_readonly = self.instance and self.instance.is_read_only()
            form.fields['amount'] = forms.IntegerField(min_value=0,
                                                       initial=0,
                                                       widget=GBtoMBWidget({'readonly': is_readonly}))

    def clean(self):
        super(PackageComponentInlineFormset, self).clean()
        if any(self._errors):
            return
        # Check whether all required components were filled and are not marked for deletion.
        marked_for_delete = []
        filled = []
        for comp in self.cleaned_data:
            if not comp:
                continue
            elif comp.get('DELETE', False):
                marked_for_delete.append(comp['type'])
            filled.append(comp['type'])
        duplicates = [item for item, count in collections.Counter(filled).items() if count > 1]
        if duplicates:
            raise forms.ValidationError(_('One or more items are duplicated: %s') % ', '.join(duplicates))

        for t in models.PackageTemplate.get_required_component_types():
            if t not in filled:
                raise forms.ValidationError(_('%s component must be specified.') % t.capitalize())
            elif t in marked_for_delete:
                raise forms.ValidationError(_('%s component is required and cannot be deleted.') % t.capitalize())


class PackageComponentInline(admin.TabularInline):
    form = PackageComponentForm
    formset = PackageComponentInlineFormset
    model = models.PackageComponent
    extra = 0

    def get_extra(self, request, obj=None, **kwargs):
        if obj:
            return super(PackageComponentInline, self).get_extra(request, obj, **kwargs)
        # On creation number of inlines must be equal to number of required components
        return len(models.PackageTemplate.get_required_component_types())


class PackageTemplateAdmin(admin.ModelAdmin):
    # WIKI: https://opennode.atlassian.net/wiki/display/WD/Shared+OpenStack+Provider+Management#SharedOpenStackProviderManagement-VPCPackagetemplatemanagement
    inlines = [PackageComponentInline]
    package_dependant_fields = ('name', 'category', 'service_settings')
    fields = package_dependant_fields + ('archived', 'icon_url', 'description', 'product_code', 'article_code')
    list_display = ('name', 'uuid', 'service_settings', 'price', 'archived', 'monthly_price', 'category')
    list_filter = ('service_settings',)
    search_fields = ('name', 'uuid')

    def get_readonly_fields(self, request, obj=None):
        if obj and obj.is_read_only():
            return self.package_dependant_fields
        else:
            return super(PackageTemplateAdmin, self).get_readonly_fields(request, obj)

    def get_form(self, request, obj=None, **kwargs):
        form = super(PackageTemplateAdmin, self).get_form(request, obj, **kwargs)

        if 'service_settings' in form.base_fields:
            form.base_fields['service_settings'].queryset = structure_models.ServiceSettings.objects.filter(shared=True)
        return form

    def save_related(self, request, form, formsets, change):
        super(PackageTemplateAdmin, self).save_related(request, form, formsets, change)
        utils.sync_price_list_item(form.instance)


class OpenStackPackageAdmin(admin.ModelAdmin):
    list_display = ('template', 'tenant', 'service_settings')
    list_filter = ('template',)


admin.site.register(models.PackageTemplate, PackageTemplateAdmin)
admin.site.register(models.OpenStackPackage, OpenStackPackageAdmin)



from __future__ import unicode_literals

from nodeconductor.core import NodeConductorExtension


class InvoicesExtension(NodeConductorExtension):
    class Settings:
        # wiki: https://opennode.atlassian.net/wiki/display/WD/Assembly+plugin+configuration
        INVOICES = {
            'ISSUER_DETAILS': {
                'company': 'OpenNode',
                'address': 'Lille 4-205',
                'country': 'Estonia',
                'email': 'info@opennodecloud.com',
                'postal': '80041',
                'phone': {
                    'country_code': '372',
                    'national_number': '5555555',
                },
                'bank': 'Estonian Bank',
                'account': '123456789',
                'vat_code': 'EE123456789',
                'country_code': 'EE',
            },
            # How many days are given to pay for created invoice
            'PAYMENT_INTERVAL': 30,
            'COMPANY_TYPES': (
                'Ministry',
                'Private company',
                'Public company',
                'Government owned company',
            ),
            'ENABLE_ACCOUNTING_START_DATE': False,
            'INVOICE_REPORTING': {
                'ENABLE': False,
                'EMAIL': 'accounting@waldur.example.com',
                'CSV_PARAMS': {
                    'delimiter': str(';'),
                },
                'USE_SAF': False,
                'SERIALIZER_EXTRA_KWARGS': {
                    'start': {
                        'format': '%d.%m.%Y',
                    },
                    'end': {
                        'format': '%d.%m.%Y',
                    }
                },
                'SAF_PARAMS': {
                    'RMAKSULIPP': '20%',
                    'ARTPROJEKT': 'PROJEKT',
                }
            },
        }

    @staticmethod
    def django_app():
        return 'nodeconductor_assembly_waldur.invoices'

    @staticmethod
    def rest_urls():
        from .urls import register_in
        return register_in

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def celery_tasks():
        from celery.schedules import crontab
        return {
            'waldur-create-invoices': {
                'task': 'invoices.create_monthly_invoices',
                'schedule': crontab(minute=0, hour=0, day_of_month='1'),
                'args': (),
            },
        }



from django.apps import AppConfig
from django.db.models import signals


class SupportConfig(AppConfig):
    name = 'nodeconductor_assembly_waldur.support'
    verbose_name = 'HelpDesk'

    def ready(self):
        from . import handlers

        Issue = self.get_model('Issue')
        Offering = self.get_model('Offering')
        Comment = self.get_model('Comment')

        signals.post_save.connect(
            handlers.log_issue_save,
            sender=Issue,
            dispatch_uid='nodeconductor_assembly_waldur.support.handlers.log_issue_save',
        )

        signals.post_delete.connect(
            handlers.log_issue_delete,
            sender=Issue,
            dispatch_uid='nodeconductor_assembly_waldur.support.handlers.log_issue_delete',
        )

        signals.post_save.connect(
            handlers.log_offering_created,
            sender=Offering,
            dispatch_uid='nodeconductor_assembly_waldur.support.handlers.log_offering_created',
        )

        signals.pre_delete.connect(
            handlers.log_offering_deleted,
            sender=Offering,
            dispatch_uid='nodeconductor_assembly_waldur.support.handlers.log_offering_deleted',
        )

        signals.post_save.connect(
            handlers.log_offering_state_changed,
            sender=Offering,
            dispatch_uid='nodeconductor_assembly_waldur.support.handlers.log_offering_state_changed',
        )

        signals.post_save.connect(
            handlers.send_comment_added_notification,
            sender=Comment,
            dispatch_uid='nodeconductor_assembly_waldur.support.handlers.send_comment_added_notification'
        )

        signals.post_save.connect(
            handlers.send_issue_updated_notification,
            sender=Issue,
            dispatch_uid='nodeconductor_assembly_waldur.support.handlers.send_issue_updated_notification'
        )




from django.utils import timezone

from nodeconductor_assembly_waldur.invoices import registrators


def add_new_allocation_to_invoice(sender, instance, created=False, **kwargs):
    if not created:
        return

    registrators.RegistrationManager.register(instance, timezone.now())


def terminate_invoice_when_allocation_deleted(sender, instance, **kwargs):
    registrators.RegistrationManager.terminate(instance, timezone.now())


def terminate_invoice_when_allocation_cancelled(sender, instance, created=False, **kwargs):
    if created:
        return

    if instance.tracker.has_changed('is_active') and not instance.is_active:
        registrators.RegistrationManager.terminate(instance, timezone.now())


def update_invoice_item_on_allocation_usage_update(sender, instance, created=False, **kwargs):
    if created:
        return

    allocation = instance

    has_changed = (allocation.tracker.has_changed('cpu_usage') or
                   allocation.tracker.has_changed('gpu_usage') or
                   allocation.tracker.has_changed('ram_usage'))
    if not has_changed:
        return

    invoice_item = registrators.RegistrationManager.get_item(allocation)
    if not invoice_item:
        return
    registrator = registrators.RegistrationManager.get_registrator(allocation)
    package = registrator.get_package(allocation)
    if package:
        invoice_item.unit_price = registrator.get_price(allocation, package)
        invoice_item.save(update_fields=['unit_price'])



import logging
import math

from nodeconductor_assembly_waldur.invoices import registrators
from nodeconductor_assembly_waldur.invoices import models as invoice_models

from waldur_slurm import models as slurm_models
from . import models

logger = logging.getLogger(__name__)


class AllocationRegistrator(registrators.BaseRegistrator):
    def get_sources(self, customer):
        return slurm_models.Allocation.objects.filter(service_project_link__project__customer=customer).distinct()

    def has_sources(self, customer):
        return self.get_sources(customer).exists()

    def get_customer(self, source):
        return source.service_project_link.project.customer

    def _find_item(self, source, now):
        allocation = source
        result = invoice_models.GenericInvoiceItem.objects.filter(
            scope=allocation,
            invoice__customer=allocation.service_project_link.project.customer,
            invoice__state=invoice_models.Invoice.States.PENDING,
            invoice__year=now.year,
            invoice__month=now.month,
        ).first()
        return result

    def _create_item(self, source, invoice, start, end):
        allocation = source
        package = self.get_package(allocation)
        if package:
            return invoice_models.GenericInvoiceItem.objects.create(
                scope=allocation,
                project=source.service_project_link.project,
                unit_price=self.get_price(allocation, package),
                unit=invoice_models.GenericInvoiceItem.Units.QUANTITY,
                quantity=1,
                product_code=package.product_code,
                article_code=package.article_code,
                invoice=invoice,
                start=start,
                end=end,
            )

    def get_package(self, allocation):
        service_settings = allocation.service_project_link.service.settings
        try:
            return models.SlurmPackage.objects.get(service_settings=service_settings)
        except models.SlurmPackage.DoesNotExist:
            logger.debug('Skip SLURM invoice item because pricing package'
                         ' for service settings %s is not defined.', service_settings)

    def get_price(self, allocation, package):
        minutes_in_hour = 60
        mb_in_gb = 1024
        cpu_price = int(math.ceil(1.0 * allocation.cpu_usage / minutes_in_hour)) * package.cpu_price
        gpu_price = int(math.ceil(1.0 * allocation.gpu_usage / minutes_in_hour)) * package.gpu_price
        ram_price = int(math.ceil(1.0 * allocation.ram_usage / mb_in_gb)) * package.ram_price
        return cpu_price + gpu_price + ram_price

    def get_details(self, source):
        return {
            'cpu_usage': source.cpu_usage,
            'gpu_usage': source.gpu_usage,
            'ram_usage': source.ram_usage,
        }
